"""Translation from HCSP to Isabelle."""

from ss2hcsp.hcsp import expr
from ss2hcsp.hcsp import hcsp


def getVariableTable(vars):
    """Give one-character name to each variable.

    vars - list(str): list of variable names.

    Returns a dictionary mapping variable names to Isabelle
    definitions. Also return the Isabelle code containing variable
    definitions.

    """
    if len(vars) > 26:
        raise AssertionError('Translate to Isabelle: too many variables')

    init_code = ""
    mapping = dict()
    for i, var in enumerate(sorted(list(vars))):
        camelCase = var[0].upper() + var[1:].lower()
        char = chr(97 + i)  # starting from 'a'
        init_code += "definition %s :: char where \"%s = CHR ''%s''\"\n" % (camelCase, camelCase, char)
        mapping[var] = camelCase

    return init_code, mapping

code_pattern = """
theory %s
    imports HHLProver.HHLProver
begin

text \<open>Variables\<close>

%s

text \<open>Processes\<close>

%s

end
"""

def translate_isabelle(process, name):
    """Translate an HCSP process to Isabelle code.

    process - list(HCSPInfo): the parallel process to be translated.
    name - str: process name, to be used as name of the Isabelle file).

    """
    # First, collect all variables
    vars = set()
    for info in process:
        vars = vars.union(info.hp.get_vars())

    init_code, mapping = getVariableTable(vars)

    def trans_expr(e):
        if isinstance(e, expr.AVar):
            return "s %s" % mapping[e.name]
        elif isinstance(e, expr.AConst):
            return str(e.value)
        elif isinstance(e, expr.OpExpr):
            if len(e.exprs) == 1:
                res = trans_expr(e.exprs[0])
                if e.exprs[0].priority() < e.priority():
                    res = "(" + res + ")"
                return "-" + res
            else:
                r1, r2 = trans_expr(e.exprs[0]), trans_expr(e.exprs[1])
                if e.exprs[0].priority() < e.priority():
                    r1 = "(" + r1 + ")"
                if e.exprs[1].priority() < e.priority():
                    r2 = "(" + r2 + ")"
                return r1 + " " + e.op + " " + r2
        elif isinstance(e, expr.BConst):
            return "True" if e.value else "False"
        elif isinstance(e, expr.RelExpr):
            return "%s %s %s" % (trans_expr(e.expr1), e.op, trans_expr(e.expr2))
        elif isinstance(e, expr.FunExpr):
            if e.fun_name in ('sqrt',):
                return "%s(%s)" % (e.fun_name, trans_expr(e.exprs[0]))
            else:
                print(e)
                raise NotImplementedError
        elif isinstance(e, expr.LogicExpr):
            if e.op == '||':
                return "%s \<or> %s" % (trans_expr(e.expr1), trans_expr(e.expr2))
            elif e.op == '&&':
                return "%s \<and> %s" % (trans_expr(e.expr1), trans_expr(e.expr2))
            else:
                raise NotImplementedError
        else:
            print(e, type(e))
            raise NotImplementedError

    def trans_lambda_expr(e):
        if not e.get_vars():
            return "(\<lambda>_. %s)" % trans_expr(e)
        else:
            return "(\<lambda>s. %s)" % trans_expr(e)

    def trans_ode_pair(v, eq):
        return "%s := %s" % (mapping[v], trans_lambda_expr(eq))

    def trans_io_comm(io, proc):
        if io.type == 'output_channel':
            io_str = "''%s''[!]%s" % (io.ch_name, trans_lambda_expr(io.expr))
        elif io.type == 'input_channel':
            io_str = "''%s''[?]%s" % (io.ch_name, mapping[io.var_name.name])
        else:
            raise NotImplementedError
        return "(%s, %s)" % (io_str, trans(proc))

    def trans(proc):
        if proc.type == 'skip':
            return "Skip"
        elif proc.type == 'assign':
            return "%s ::= %s" % (mapping[proc.var_name.name], trans_lambda_expr(proc.expr))
        elif proc.type == 'sequence':
            return ";\n".join(trans(p) for p in proc.hps)
        elif proc.type == 'loop':
            body = trans(proc.hp)
            indent_body = '\n'.join('  ' + line for line in body.split('\n'))
            return "Rep (\n%s\n)" % indent_body
        elif proc.type == 'output_channel':
            return "Cm (''%s''[!]%s)" % (proc.ch_name, trans_lambda_expr(proc.expr))
        elif proc.type == 'input_channel':
            return "Cm (''%s''[?]%s)" % (proc.ch_name, mapping[proc.var_name.name])
        elif proc.type == 'ite':
            cond, body = proc.if_hps[0]
            body = trans(body)
            indent_body = '\n'.join('  ' + line for line in body.split('\n'))
            if len(proc.if_hps) > 1:
                else_body = trans(hcsp.ITE(proc.if_hps[1:], proc.else_hp))
            else:
                else_body = trans(proc.else_hp)
            indent_else = '\n'.join('  ' + line for line in body.split('\n'))
            return "IF (%s) THEN\n%s\nELSE\n%s\nFI" % (
                trans_lambda_expr(cond), indent_body, indent_else)
        elif proc.type == 'wait':
            return "Wait (%s)" % trans_lambda_expr(proc.delay)
        elif proc.type == 'ode':
            return "Cont (ODE ((\<lambda>_ _. 0)(%s))) (%s)" % (
                ', '.join(trans_ode_pair(v, eq) for v, eq in proc.eqs), trans_lambda_expr(proc.constraint))
        elif proc.type == 'ode_comm':
            return "Interrupt (ODE ((\<lambda>_ _. 0)(%s))) (%s)\n[%s]" % (
                ', '.join(trans_ode_pair(v, eq) for v, eq in proc.eqs),
                trans_lambda_expr(proc.constraint),
                ', '.join(trans_io_comm(io, proc) for io, proc in proc.io_comms))
        else:
            print(proc.type, proc)
            raise NotImplementedError

    proc_code_list = []
    for info in process:
        proc_code = "definition %s :: proc where\n  \"%s =" % (info.name, info.name)
        for line in trans(info.hp).split('\n'):
            proc_code += "\n    " + line
        proc_code += "\"\n"
        proc_code_list.append(proc_code)

    code = code_pattern % (name, init_code, '\n'.join(proc_code for proc_code in proc_code_list))

    return code



"""Discrete PID Controller"""

from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.expr import AVar, AConst, RelExpr, OpExpr
from ss2hcsp.hcsp import hcsp as hp


class DiscretePID(SL_Block):
    """Block for unit delay."""
    def __init__(self, name, controller="PI", st=-1, pid=(), init_value=0, saturation=(), kb=0):
        super(DiscretePID, self).__init__()
        self.name = name
        self.type = "discrete_PID_controller"
        self.is_continuous = False
        self.num_src = 1
        self.num_dest = 1
        self.src_lines = [[]]
        self.dest_lines = [None]

        self.controller = controller
        assert st > 0
        self.st = st
        assert len(pid) == 3
        self.pid = pid
        self.init_value = init_value
        assert len(saturation) == 2 and saturation[0] < saturation[1]
        self.saturation = saturation
        self.kb = kb

    def __str__(self):
        return "%s: DiscretePID[controller = %s, in = %s, out = %s, st = %s]" % \
               (self.name, str(self.controller), str(self.dest_lines), str(self.src_lines), str(self.st))

    def __repr__(self):
        return str(self)

    def get_init_hps(self):
        return [hp.Assign(var_name="Integrator_DSTATE", expr=AConst(self.init_value))]

    def get_output_hp(self):
        in_var = self.dest_lines[0].name
        out_var = self.src_lines[0][0].name

        step0 = hp.Assign(var_name="rtb_Sum", expr=AVar(in_var))
        step1 = hp.Assign(var_name="rtb_IntegralGain", expr=OpExpr("*", AVar("rtb_Sum"), AConst(self.pid[1])))
        step2 = hp.Assign(var_name="rtb_Sum", expr=OpExpr("*", AVar("rtb_Sum"), AConst(self.pid[0])))
        step3 = hp.Assign(var_name="rtb_SumFdbk", expr=OpExpr("+", AVar("rtb_Sum"), AVar("Integrator_DSTATE")))
        step4 = hp.ITE(if_hps=[(RelExpr(">", AVar("rtb_SumFdbk"), AConst(self.saturation[1])),
                                hp.Assign(var_name="rtb_SumFdbk_0", expr=AConst(self.saturation[1]))),
                               (RelExpr("<", AVar("rtb_SumFdbk"), AConst(self.saturation[0])),
                                hp.Assign(var_name="rtb_SumFdbk_0", expr=AConst(self.saturation[0])))],
                       else_hp=hp.Assign(var_name="rtb_SumFdbk_0", expr=AVar("rtb_SumFdbk"))
                       )
        expr0 = OpExpr("*", OpExpr("-", AVar("rtb_SumFdbk_0"), AVar("rtb_SumFdbk")), AConst(self.kb))
        expr1 = OpExpr("*", OpExpr("+", expr0, AVar("rtb_IntegralGain")), AConst(self.st))
        step5 = hp.Assign(var_name="rtb_IntegralGain", expr=OpExpr("+", expr1, AVar("Integrator_DSTATE")))
        step6 = hp.Assign(var_name="rtb_Sum", expr=OpExpr("+", AVar("rtb_Sum"), AVar("rtb_IntegralGain")))
        step7 = hp.ITE(if_hps=[(RelExpr(">", AVar("rtb_Sum"), AConst(self.saturation[1])),
                                hp.Assign(var_name=out_var, expr=AConst(self.saturation[1]))),
                               (RelExpr("<", AVar("rtb_Sum"), AConst(self.saturation[0])),
                                hp.Assign(var_name=out_var, expr=AConst(self.saturation[0])))],
                       else_hp=hp.Assign(var_name=out_var, expr=AVar("rtb_Sum")))
        step8 = hp.Assign(var_name="Integrator_DSTATE", expr=AVar("rtb_IntegralGain"))

        time_cond = RelExpr("==", OpExpr("%", AVar("t"), AConst(self.st)), AConst(0))

        return hp.Condition(cond=time_cond,
                            hp=hp.Sequence(step0, step1, step2, step3, step4, step5, step6, step7, step8))



"""Parser for expressions."""

from lark import Lark, Transformer, v_args, exceptions
from lark.tree import Meta
from ss2hcsp.hcsp import expr
from ss2hcsp.hcsp import assertion, label
from ss2hcsp.hcsp import hcsp
from ss2hcsp.hcsp import module
from decimal import Decimal


grammar = r"""
    ?lname: CNAME -> var_expr
        | CNAME "[" expr "]" -> array_idx_expr
        | CNAME "[" expr "]" "[" expr "]" -> array_idx_expr2
        | CNAME "[" expr "]" "[" expr "]""[" expr "]" -> array_idx_expr3
        | lname "." CNAME -> field_expr
        | lname "." CNAME "[" expr "]" -> field_array_idx

    ?atom_expr: lname
        | SIGNED_NUMBER -> num_expr
        | ESCAPED_STRING -> string_expr
        | "[" "]" -> empty_list
        | "[" expr ("," expr)* "]" -> literal_list
        | "{" "}" -> empty_dict
        | "{" CNAME ":" expr ("," CNAME ":" expr)* "}" -> literal_dict
        | "min" "(" expr "," expr ")" -> min_expr
        | "max" "(" expr "," expr ")" -> max_expr
        | "gcd" "(" expr ("," expr)+ ")" -> gcd_expr
        | "true" -> true_cond
        | "false" -> false_cond
        | CNAME "(" (expr)? ("," expr)* ")" -> fun_expr
        | "(" expr ")"

    ?power_expr: power_expr "^" atom_expr -> power_expr    // priority 85
        | atom_expr

    ?uminus: "-" uminus -> uminus | power_expr              // Unary minus: priority 80

    ?times_expr: times_expr "*" uminus -> times_expr       // priority 70
        | times_expr "/" uminus -> divide_expr
        | times_expr "%" uminus -> mod_expr
        | uminus

    ?plus_expr: plus_expr "+" times_expr -> plus_expr      // priority 65
        | plus_expr "-" times_expr -> minus_expr
        | times_expr

    ?if_expr: "if" expr "then" if_expr "else" if_expr       // priority 40
        | plus_expr

    ?rel_expr: if_expr "==" if_expr -> eq_cond         // priority 50
        | if_expr "!=" if_expr -> ineq_cond
        | if_expr "<=" if_expr -> less_eq_cond
        | if_expr "<" if_expr -> less_cond
        | if_expr ">=" if_expr -> greater_eq_cond
        | if_expr ">" if_expr -> greater_cond
        | if_expr

    ?neg_expr: "!" neg_expr -> not_cond          // priority 40
        | rel_expr

    ?conj: neg_expr "&&" conj | neg_expr         // Conjunction: priority 35

    ?disj: conj "||" disj | conj                 // Disjunction: priority 30

    ?equiv: disj "<->" equiv | disj             // Equivalent: priority 25

    ?imp: equiv "->" imp | equiv                  // Implies: priority 20

    ?quant: "\\exists" CNAME "." quant                         -> exists_expr  // priority 10
        | "\\exists" "{" CNAME ("," CNAME)+ "}" "." quant      -> exists_expr
        | "\\forall" CNAME "." quant                       -> forall_expr
        | "\\forall" "{" CNAME ("," CNAME)+ "}" "." quant  -> forall_expr
        | imp

    ?expr: quant

    ?comm_cmd: CNAME "?" lname -> input_cmd
        | CNAME "[" expr "]" "?" lname -> input_cmd
        | CNAME "[" expr "]" "[" expr "]" "?" lname -> input_cmd
        | CNAME "?" -> input_none_cmd
        | CNAME "[" expr "]" "?" -> input_none_cmd
        | CNAME "[" expr "]" "[" expr "]" "?" -> input_none_cmd
        | CNAME "!" expr -> output_cmd
        | CNAME "[" expr "]" "!" expr -> output_cmd
        | CNAME "[" expr "]" "[" expr "]" "!" expr -> output_cmd
        | CNAME "!" -> output_none_cmd
        | CNAME "[" expr "]" "!" -> output_none_cmd
        | CNAME "[" expr "]" "[" expr "]" "!" -> output_none_cmd

    ?ode_seq: CNAME "=" expr ("," CNAME "=" expr)*

    ?interrupt: comm_cmd "-->" cmd ("," comm_cmd "-->" cmd)*

    ?atom_cmd: "@" CNAME ";" -> var_cmd
        | "skip" ";" -> skip_cmd
        | "wait" "(" expr ")" ";" -> wait_cmd
        | lname ":=" expr ";" -> assign_cmd
        | "(" lname ("," lname)* ")" ":=" expr ";" -> multi_assign_cmd
        | lname ":=" "*" "(" expr ")" ";" -> random_assign_cmd
        | "assert" "(" expr ("," expr)* ")" ";" -> assert_cmd
        | "test" "(" expr ("," expr)* ")" ";" -> test_cmd
        | "log" "(" expr ("," expr)* ")" ";" -> log_cmd
        | comm_cmd ";"
        | "{" cmd "}" "*" maybe_loop_invariant -> repeat_cmd
        | "{" cmd "}" "*" "(" expr ")" maybe_loop_invariant -> repeat_cond_cmd
        | "{" ode_seq "&" expr "}" maybe_ode_invariant -> ode
        | "{" "&" expr "}" "|>" "[]" "(" interrupt ")" maybe_ode_invariant -> ode_comm_const
        | "{" ode_seq "&" expr "}" "|>" "[]" "(" interrupt ")" maybe_ode_invariant -> ode_comm
        | "if" "(" expr ")" "{" cmd "}" ("else" "if" "(" expr ")" "{" cmd "}")* ("else" "{" cmd "}")? -> ite_cmd
        | "{" cmd "}" -> paren_cmd

    ?maybe_loop_invariant: ("invariant" ord_assertion+ ";")? -> maybe_loop_invariant

    ?ord_assertion: "[" expr "]" maybe_proof_methods          -> ord_assertion
    
    ?maybe_proof_methods: ("{{" proof_method ("," proof_method)* "}}")?    ->maybe_proof_methods
    ?proof_method: (label ":")? method   -> proof_method 
    
    ?label_category: "init"     -> label_categ_init
        | "maintain"         -> label_categ_maintain

    ?categ_label: label_category     -> categ_label

    ?atom_label: INT
        | "execute"          -> atom_label_execute
        | "skip"             -> atom_label_skip

    ?branch_label: atom_label                 -> atom_label
        | branch_label ("." branch_label)+           -> seq_label
        | atom_label "(" branch_label ")"     -> nest_label

    ?label: categ_label                  
        | branch_label
        | categ_label branch_label       -> comp_label
    
    ?method: "z3"        -> method_z3
      | "wolfram"        -> method_wolfram

    ?maybe_ode_invariant: ("invariant" ode_invariant+ ";")? -> maybe_ode_invariant

    ?ode_invariant: "[" expr "]" ("{" ode_rule expr? "}")? maybe_proof_methods -> ode_invariant
        | "ghost" CNAME -> ghost_intro
        | "ghost" "(" CNAME "=" expr ")" -> ghost_intro_eq

    ?ode_rule: "di" -> ode_rule_di
      | "dbx" -> ode_rule_dbx
      | "bc" -> ode_rule_bc
      | "dw" -> ode_rule_dw
      | "sln" -> ode_rule_sln

    ?ichoice_cmd: atom_cmd ("++" atom_cmd)*  // Priority: 80

    ?seq_cmd: ichoice_cmd*  // Priority: 70

    ?select_cmd: seq_cmd | comm_cmd "-->" seq_cmd ("$" comm_cmd "-->" seq_cmd)*  // Priority 50

    ?cmd: select_cmd

    ?function_decl: "function" CNAME "(" CNAME ("," CNAME)* ")" "=" expr ";"

    ?hoare_pre : "pre" (ord_assertion)* ";" -> hoare_pre
    ?hoare_post : "post" (ord_assertion)* ";" -> hoare_post
    ?hoare_triple: (function_decl)* hoare_pre cmd hoare_post

    ?procedure: "procedure" CNAME "begin" cmd "end"

    ?parallel_cmd: cmd ("||" cmd)*   // Priority 30, outermost level only

    ?module_sig: CNAME "(" (CNAME | "$" CNAME) ("," (CNAME | "$" CNAME))* ")"  -> module_sig
        | CNAME "(" ")"                            -> module_sig

    ?module_output: "output" CNAME ("," CNAME)* ";"    -> module_output
    
    ?module: "module" module_sig ":" (module_output)* (procedure)* "begin" cmd "end" "endmodule"
    
    ?module_arg: CNAME ("[" INT "]")*   -> module_arg_channel
        | "$" expr    -> module_arg_expr

    ?module_args: CNAME "(" module_arg ("," module_arg)* ")"  -> module_args
        | CNAME "(" ")"                            -> module_args

    ?module_inst: module_args    -> module_inst_noname
        | CNAME "=" module_args  -> module_inst

    ?system: "system" module_inst ("||" module_inst)* "endsystem"  -> system

    ?import: "import" CNAME   -> hcsp_import

    ?decls: "%type: module" (module | import | system)* -> decls

    %import common.CNAME
    %import common.WS
    %import common.INT
    %import common.DECIMAL
    %import common.NUMBER
    %import common.SIGNED_NUMBER
    %import common.ESCAPED_STRING

    COMMENT: /#.*/

    %ignore WS
    %ignore COMMENT
"""

def _vargs_meta_inline(f, _data, children, meta):
    return f(meta, *children)

@v_args(wrapper=_vargs_meta_inline)
class HPTransformer(Transformer):
    def __init__(self):
        pass

    def var_expr(self, meta, s):
        return expr.AVar(str(s), meta=meta)

    def num_expr(self, meta, v):
        return expr.AConst(Decimal(str(v)) if '.' in v or 'e' in v else int(v), meta=meta)

    def string_expr(self, meta, s):
        return expr.AConst(str(s)[1:-1], meta=meta)  # remove quotes

    def empty_list(self, meta):
        return expr.ListExpr(meta=meta)

    def literal_list(self, meta, *args):
        if all(isinstance(arg, expr.AConst) for arg in args):
            return expr.AConst(list(arg.value for arg in args), meta=meta)
        else:
            return expr.ListExpr(*args, meta=meta)

    def empty_dict(self, meta):
        return expr.DictExpr(meta=meta)

    def literal_dict(self, meta, *args):
        # args should contain 2*n elements, which are key-value pairs
        assert len(args) >= 2 and len(args) % 2 == 0
        pairs = []
        for i in range(0,len(args),2):
            pairs.append((str(args[i]), args[i+1]))
        return expr.DictExpr(*pairs, meta=meta)

    def array_idx_expr(self, meta, a, i):
        return expr.ArrayIdxExpr(expr.AVar(str(a)), i, meta=meta)
    
    def array_idx_expr2(self, meta, a, i, j):
        return expr.ArrayIdxExpr(expr.ArrayIdxExpr(expr.AVar(str(a)), i, meta=meta), j, meta=meta)

    def array_idx_expr3(self, meta, a, i, j, k):
        return expr.ArrayIdxExpr(expr.ArrayIdxExpr(expr.ArrayIdxExpr(expr.AVar(str(a)), i, meta=meta), j, meta=meta), k, meta=meta)

    def field_array_idx(self, meta, e, field, i):
        return expr.ArrayIdxExpr(expr.FieldNameExpr(e, field, meta=meta), i, meta=meta)

    def field_expr(self, meta, e, field):
        return expr.FieldNameExpr(e, field, meta=meta)

    def if_expr(self, meta, cond, e1, e2):
        return expr.IfExpr(cond, e1, e2, meta=meta)

    def exists_expr(self, meta, *args):
        # The last arg is the expression, other args are variables.
        assert len(args) >= 2
        e = args[-1]
        if len(args) == 2:
            return expr.ExistsExpr(str(args[0]), e, meta=meta)
        else:
            return expr.ExistsExpr(list(str(arg) for arg in args[:-1]), e, meta=meta)

    def forall_expr(self, meta, *args):
        # The last arg is the expression, other args are variables.
        assert len(args) >= 2
        e = args[-1]
        if len(args) == 2:
            return expr.ForAllExpr(str(args[0]), e, meta=meta)
        else:
            return expr.ForAllExpr(list(str(arg) for arg in args[:-1]), e, meta=meta)

    def plus_expr(self, meta, e1, e2):
        return expr.OpExpr("+", e1, e2, meta=meta)

    def minus_expr(self, meta, e1, e2):
        return expr.OpExpr("-", e1, e2, meta=meta)

    def uminus(self, meta, e):
        return expr.OpExpr("-", e, meta=meta)

    def times_expr(self, meta, e1, e2):
        return expr.OpExpr("*", e1, e2, meta=meta)

    def divide_expr(self, meta, e1, e2):
        return expr.OpExpr("/", e1, e2, meta=meta)

    def power_expr(self, meta, e1, e2):
        return expr.OpExpr("^", e1, e2, meta=meta)

    def min_expr(self, meta, e1, e2):
        return expr.FunExpr("min", [e1, e2], meta=meta)

    def max_expr(self, meta, e1, e2):
        return expr.FunExpr("max", [e1, e2], meta=meta)

    def mod_expr(self, meta, e1, e2):
        return expr.OpExpr("%", e1, e2, meta=meta)

    def gcd_expr(self, meta, *exprs):
        return expr.FunExpr(fun_name="gcd", exprs=exprs, meta=meta)

    def fun_expr(self, meta, fun_name, *exprs):
        return expr.FunExpr(str(fun_name), exprs, meta=meta)

    def eq_cond(self, meta, e1, e2):
        return expr.RelExpr("==", e1, e2, meta=meta)

    def ineq_cond(self, meta, e1, e2):
        return expr.RelExpr("!=", e1, e2, meta=meta)

    def less_eq_cond(self, meta, e1, e2):
        return expr.RelExpr("<=", e1, e2, meta=meta)

    def less_cond(self, meta, e1, e2):
        return expr.RelExpr("<", e1, e2, meta=meta)

    def greater_eq_cond(self, meta, e1, e2):
        return expr.RelExpr(">=", e1, e2, meta=meta)

    def greater_cond(self, meta, e1, e2):
        return expr.RelExpr(">", e1, e2, meta=meta)

    def not_cond(self, meta, e):
        return expr.LogicExpr("!", e, meta=meta)

    def true_cond(self, meta):
        return expr.BConst(True, meta=meta)

    def false_cond(self, meta):
        return expr.BConst(False, meta=meta)

    def conj(self, meta, b1, b2):
        return expr.LogicExpr("&&", b1, b2, meta=meta)

    def disj(self, meta, b1, b2):
        return expr.LogicExpr("||", b1, b2, meta=meta)

    def imp(self, meta, b1, b2):
        return expr.LogicExpr("->", b1, b2, meta=meta)

    def equiv(self, meta, b1, b2):
        return expr.LogicExpr("<->", b1, b2, meta=meta)

    def var_cmd(self, meta, name):
        return hcsp.Var(str(name), meta=meta)

    def skip_cmd(self, meta):
        return hcsp.Skip(meta=meta)

    def wait_cmd(self, meta, delay):
        return hcsp.Wait(delay, meta=meta)

    def assign_cmd(self, meta, var, expr):
        return hcsp.Assign(var, expr, meta=meta)

    def multi_assign_cmd(self, meta, *args):
        return hcsp.Assign(args[:-1], args[-1], meta=meta)

    def random_assign_cmd(self, meta, var, cond):
        return hcsp.RandomAssign(var, cond, meta=meta)

    def assert_cmd(self, meta, *args):
        # First argument is the assert condition. The remaining ones
        # are error messages. 
        bexpr, msgs = args[0], args[1:]
        return hcsp.Assert(bexpr, msgs, meta=meta)

    def test_cmd(self, meta, *args):
        # First argument is the test condition. The remaining ones
        # are warning messages.
        bexpr, msgs = args[0], args[1:]
        return hcsp.Test(bexpr, msgs, meta=meta)

    def log_cmd(self, meta, *args):
        return hcsp.Log(args[0], exprs=args[1:], meta=meta)

    def seq_cmd(self, meta, *args):
        if len(args) == 0:
            return hcsp.Skip(meta=meta)
        elif len(args) == 1:
            return args[0]
        else:
            return hcsp.Sequence(*args, meta=meta)

    def input_cmd(self, meta, *args):
        # First argument is channel name, last argument is variable name.
        # Middle arguments are args to channel name.
        ch_name, ch_args, var_name = args[0], args[1:-1], args[-1]
        return hcsp.InputChannel(hcsp.Channel(str(ch_name), ch_args, meta=meta), var_name, meta=meta)

    def input_none_cmd(self, meta, *args):
        # First argument is channel name, remaining arguments are its args.
        ch_name, ch_args = args[0], args[1:]
        return hcsp.InputChannel(hcsp.Channel(str(ch_name), ch_args, meta=meta), meta=meta)

    def output_cmd(self, meta, *args):
        # First argument is channel name, last argument is variable name.
        # Middle arguments are args to channel name.
        ch_name, ch_args, expr = args[0], args[1:-1], args[-1]
        return hcsp.OutputChannel(hcsp.Channel(str(ch_name), ch_args, meta=meta), expr, meta=meta)

    def output_none_cmd(self, meta, *args):
        # First argument is channel name, remaining arguments are its args.
        ch_name, ch_args = args[0], args[1:]
        return hcsp.OutputChannel(hcsp.Channel(str(ch_name), ch_args, meta=meta), meta=meta)

    def ord_assertion(self, meta, expr, proof_methods):
        return assertion.OrdinaryAssertion(expr=expr, proof_methods=proof_methods, meta=meta)

    def maybe_loop_invariant(self, meta, *args):
        return args

    def maybe_ode_invariant(self, meta, *args):
        return args

    def ode_invariant(self, meta, *args):
        if len(args) == 2:
            return assertion.CutInvariant(expr=args[0], proof_methods=args[-1], meta=meta)
        elif len(args) == 3:
            return assertion.CutInvariant(expr=args[0], rule=args[1], proof_methods=args[-1], meta=meta)
        else:
            return assertion.CutInvariant(expr=args[0], rule=args[1], rule_arg=args[2], proof_methods=args[-1], meta=meta)
    
    def ghost_intro(self, meta, var):
        return assertion.GhostIntro(var=var, diff=None, meta=meta)
    
    def ghost_intro_eq(self, meta, var, diff):
        assert var.endswith("_dot")
        return assertion.GhostIntro(var=var[:-4], diff=diff, meta=meta)

    def ode_rule_di(self, meta): return "di"
    def ode_rule_bc(self, meta): return "bc"
    def ode_rule_dbx(self, meta): return "dbx"
    def ode_rule_dw(self, meta): return "dw"
    def ode_rule_sln(self, meta): return "sln"

    def method_z3(self, meta): return "z3"
    def method_wolfram(self, meta): return "wolfram"

    def label_categ_init(self, meta): return "init"
    def label_categ_maintain(self, meta): return "maintain"

    def atom_label_execute(self, meta): return "execute"
    def atom_label_skip(self, meta): return "skip"

    def atom_label(self, meta, value):
        return label.AtomLabel(value=value)

    def seq_label(self, meta, *args):
        return label.SequenceLabel(*args)

    def nest_label(self, meta, value, sub_label):
        return label.NestLabel(value=value, sub_label=sub_label)

    def categ_label(self, meta, categ):
        return label.CategLabel(categ=categ)

    def comp_label(self, meta, categ_label, branch_label):
        return label.CompLabel(categ_label=categ_label, branch_label=branch_label)

    def proof_method(self, meta, *args):
        assert(len(args) == 1 or len(args) == 2)
        if len(args) == 1:
            return assertion.ProofMethod(method=args[0], meta=meta)
        else:
            return assertion.ProofMethod(label=args[0], method=args[1], meta=meta)

    def maybe_proof_methods(self, meta, *args):
        if len(args) == 0:
            return assertion.ProofMethods(meta=meta)
        else:
            return assertion.ProofMethods(*args, meta=meta)
            
    def repeat_cmd(self, meta, cmd, inv):
        return hcsp.Loop(cmd, meta=meta, inv=inv)

    def repeat_cond_cmd(self, meta, cmd, cond, inv):
        return hcsp.Loop(cmd, constraint=cond, meta=meta, inv=inv)

    def ode_seq(self, meta, *args):
        res = []
        for i in range(0,len(args),2):
            assert args[i].endswith("_dot")
            res.append((str(args[i][:-4]), args[i+1]))
        return res

    def interrupt(self, meta, *args):
        res = []
        for i in range(0, len(args), 2):
            res.append((args[i], args[i+1]))
        return res

    def ode(self, meta, eqs, constraint, inv):
        return hcsp.ODE(eqs, constraint, meta=meta, inv=inv)

    def ode_comm_const(self, meta, constraint, io_comms, inv):
        return hcsp.ODE_Comm([], constraint, io_comms, meta=meta)

    def ode_comm(self, meta, eqs, constraint, io_comms, inv):
        return hcsp.ODE_Comm(eqs, constraint, io_comms, meta=meta)

    def ichoice_cmd(self, meta, *args):
        if len(args) == 1:
            return args[0]
        else:
            return hcsp.IChoice(*args, meta=meta)

    def select_cmd(self, meta, *args):
        assert len(args) % 2 == 0 and len(args) >= 4
        io_comms = []
        for i in range(0, len(args), 2):
            io_comms.append((args[i], args[i+1]))
        return hcsp.SelectComm(*io_comms, meta=meta)

    def ite_cmd(self, meta, *args):
        if_hps = []
        for i in range(0, len(args)//2):
            if_hps.append((args[2*i], args[2*i+1]))
        if len(args) % 2 == 1:
            else_hp = args[-1]
        else:
            else_hp = None
        return hcsp.ITE(if_hps, else_hp, meta=meta)

    def paren_cmd(self, meta, c):
        return c

    def procedure(self, meta, name, hp):
        return hcsp.Procedure(name, hp, meta=meta)

    def parallel_cmd(self, meta, *args):
        if len(args) == 1:
            return args[0]
        else:
            return hcsp.Parallel(*args, meta=meta)

    def function_decl(self, meta, *args):
        name = str(args[0])
        vars = list(str(var) for var in args[1:-1])
        expr = args[-1]
        return hcsp.Function(name, vars, expr)

    def hoare_pre(self, meta, *args):
        return list(args)

    def hoare_post(self, meta, *args):
        return list(args)

    def hoare_triple(self, meta, *args):
        # The last three arguments are pre, hp, and post.
        pre = args[-3]
        hp = args[-2]
        post = args[-1]

        # The other arguments are functions
        functions = dict()
        for item in args[:-3]:
            assert isinstance(item, hcsp.Function);
            functions[item.name] = item
        return hcsp.HoareTriple(pre, hp, post, functions=functions, meta=meta)

    def module_sig(self, meta, *args):
        return tuple(str(arg) for arg in args)

    def module_output(self, meta, *args):
        return tuple(str(arg) for arg in args)

    def module(self, meta, *args):
        # The first item is a tuple consisting of name and list of parameters
        # The middle items are the output sequences or procedure declarations
        # The last item is code for the module
        assert len(args) >= 2
        sig, decls, code = args[0], args[1:-1], args[-1]
        if isinstance(sig, tuple):
            name, params = sig[0], sig[1:]
        else:
            name, params = sig, tuple()
        outputs, procedures = [], []
        for decl in decls:
            if isinstance(decl, tuple):
                outputs.append(decl)
            elif isinstance(decl, hcsp.Procedure):
                procedures.append(decl)
            else:
                raise NotImplementedError
        return module.HCSPModule(name, code, params=params, outputs=outputs, procedures=procedures, meta=meta)

    def module_arg_channel(self, meta, *args):
        # First argument is channel name, remaining arguments are channel args.
        ch_name, ch_args = args[0], args[1:]
        return hcsp.Channel(str(ch_name), tuple(expr.AConst(ch_arg, meta=meta) for ch_arg in ch_args), meta=meta)

    def module_arg_expr(self, meta, expr):
        return expr

    def module_args(self, meta, *args):
        return args

    def module_inst(self, meta, name, sig):
        return module.HCSPModuleInst(name, sig[0], sig[1:], meta=meta)
    
    def module_inst_noname(self, meta, sig):
        return module.HCSPModuleInst(sig[0], sig[0], sig[1:], meta=meta)

    def hcsp_import(self, meta, filename):
        # Importing from a file
        text = module.read_file(filename + '.txt')
        if text is None:
            raise ParseFileException('File %s not found' % filename)
        try:
            # Preprocessing: just remove comments
            text_lines = text.strip().split('\n')
            text = ""

            for line in text_lines:
                comment_pos = line.find('#')
                if comment_pos != -1:
                    line = line[:comment_pos]
                if line.strip() != "":
                    text += line + '\n'

            return decls_parser.parse(text)
        except (exceptions.UnexpectedToken, exceptions.UnexpectedCharacters) as e:
            error_str = "Unable to parse\n"
            for i, line in enumerate(text.split('\n')):
                error_str += line + '\n'
                if i == e.line - 1:
                    error_str += " " * (e.column-1) + "^" + '\n'
            raise ParseFileException(error_str)

    def system(self, meta, *args):
        # Each item is a module instantiation
        return module.HCSPSystem(args, meta=meta)

    def decls(self, meta, *args):
        # A list of declarations.
        return module.HCSPDeclarations(args, meta=meta)

hp_transformer = HPTransformer()

expr_parser = Lark(grammar, start="expr", parser="lalr", transformer=hp_transformer)
hp_parser = Lark(grammar, start="parallel_cmd", parser="lalr", transformer=hp_transformer)
module_parser = Lark(grammar, start="module", parser="lalr", transformer=hp_transformer)
system_parser = Lark(grammar, start="system", parser="lalr", transformer=hp_transformer)
decls_parser = Lark(grammar, start="decls", parser="lalr", transformer=hp_transformer)

# Variants of the parsers without internal transformer, returning a Lark Tree instead of a HCSP expression.
# They allow us to get meta information about line and character numbers of the parsed code.
expr_tree_parser = Lark(grammar, start="expr", parser="lalr", propagate_positions=True)
hp_tree_parser = Lark(grammar, start="parallel_cmd", parser="lalr", propagate_positions=True)
hoare_triple_tree_parser = Lark(grammar, start="hoare_triple", parser="lalr", propagate_positions=True)
module_tree_parser = Lark(grammar, start="module", parser="lalr", propagate_positions=True)
system_tree_parser = Lark(grammar, start="system", parser="lalr", propagate_positions=True)
decls_tree_parser = Lark(grammar, start="decls", parser="lalr", propagate_positions=True)

def parse_expr_with_meta(text): return hp_transformer.transform(expr_tree_parser.parse(text))
def parse_hp_with_meta(text): return hp_transformer.transform(hp_tree_parser.parse(text))
def parse_hoare_triple_with_meta(text): return hp_transformer.transform(hoare_triple_tree_parser.parse(text))
def parse_module_with_meta(text): return hp_transformer.transform(module_tree_parser.parse(text))
def parse_system_with_meta(text): return hp_transformer.transform(system_tree_parser.parse(text))
def parse_decls_with_meta(text): return hp_transformer.transform(decls_tree_parser.parse(text))


class ParseFileException(Exception):
    def __init__(self, error_msg):
        self.error_msg = error_msg


def parse_file(text):
    """Parsing regular HCSP files.

    Input is the string of the file. Output is a list of pairs (name, hp).

    """
    text_lines = text.strip().split('\n')
    hcsp_info = []

    # First, read lines from file, each line containing ::= means the
    # start of a new program.
    lines = []
    for line in text_lines:
        comment_pos = line.find('#')
        if comment_pos != -1:
            line = line[:comment_pos]
        if line.find('::=') != -1:
            lines.append(line)
        else:
            if line != "":
                if len(lines) == 0:
                    raise ParseFileException('Unexpected line: ' + line)
                lines[-1] += line + '\n'

    infos = []

    # Now each entry in lines represent the definition of a program.
    for line in lines:
        index = line.index('::=')
        name = line[:index].strip()
        hp_text = line[index+3:].strip()

        try:
            hp = hp_parser.parse(hp_text)
        except (exceptions.UnexpectedToken, exceptions.UnexpectedCharacters) as e:
            error_str = "Unable to parse\n"
            for i, line in enumerate(hp_text.split('\n')):
                error_str += line + '\n'
                if i == e.line - 1:
                    error_str += " " * (e.column-1) + "^" + '\n'
            raise ParseFileException(error_str)

        infos.append(hcsp.HCSPInfo(name, hp))

    return infos

def parse_module_file(text):
    """Parse a file in module format.
    
    Input is the string of the file. Output is a list of pairs (name, hp).

    """
    # Preprocessing: just remove comments
    text_lines = text.strip().split('\n')
    text = ""

    for line in text_lines:
        comment_pos = line.find('#')
        if comment_pos != -1:
            line = line[:comment_pos]
        if line.strip() != "":
            text += line + '\n'

    try:
        decls = decls_parser.parse(text)
        infos = decls.generateHCSPInfo()
    except (exceptions.UnexpectedToken, exceptions.UnexpectedCharacters) as e:
        error_str = "Unable to parse\n"
        for i, line in enumerate(text.split('\n')):
            error_str += line + '\n'
            if i == e.line - 1:
                error_str += " " * (e.column-1) + "^" + '\n'
        raise ParseFileException(error_str)
    except module.ModuleException as e:
        raise ParseFileException("Module exception\n" + e.error_msg)

    return infos



from flask import Flask
from flask import request
import lark
import json
import math
import time
from pstats import Stats
import cProfile
import random

import sys
sys.path.append("..")
from ss2hcsp.hcsp import simulator
from ss2hcsp.hcsp import parser
from ss2hcsp.hcsp import pprint
from ss2hcsp.hcsp import hcsp
from ss2hcsp.server.get_port_service import get_aadl_port_service, get_simulink_port_service
from ss2hcsp.server.sequence_diagram import print_sequence_diagram

app = Flask(__name__)


@app.route('/')
def hello_world():
    return "Hello, World!"

def raise_error(err_str):
    return json.dumps({
        'error': err_str
    })

@app.route('/parse_hcsp', methods=['POST'])
def parse_hcsp():
    data = json.loads(request.get_data())
    text = data['text']

    try:
        if text.startswith('%type: module'):
            infos = parser.parse_module_file(text)
        else:
            infos = parser.parse_file(text)
    except parser.ParseFileException as e:
        return raise_error(e.error_msg)

    sim_infos = []
    hcsp_info = []
    for info in infos:
        name, hp, procs = info.name, info.hp, info.procedures
        if hp.type == 'parallel':
            if not all(sub_hp.type == 'var' for sub_hp in hp.hps):
                return raise_error("Group definition must be a parallel of variables.\n  %s" % line)

            hcsp_info.append({
                'name': name,
                'parallel': [sub_hp.name for sub_hp in hp.hps]
            })
        else:
            sim_infos.append(simulator.SimInfo(name, hp, outputs=info.outputs, procedures=info.procedures))
            lines, mapping = pprint.pprint_lines(hp, record_pos=True)
            json_procs = []
            for proc in info.procedures:
                proc_lines, proc_mapping = pprint.pprint_lines(proc.hp, record_pos=True)
                json_procs.append({
                    'name': proc.name,
                    'text': str(proc.hp),
                    'lines': proc_lines,
                    'mapping': proc_mapping
                })
            hcsp_info.append({
                'name': name,
                'text': str(hp),
                'outputs': info.outputs,
                'lines': lines,
                'mapping': mapping,
                'procedures': json_procs
            })

    warnings = simulator.check_comms(sim_infos)

    return json.dumps({
        'hcsp_info': hcsp_info,
        'warnings': warnings
    })

@app.route('/run_hcsp', methods=['POST'])
def run_hcsp():
    data = json.loads(request.get_data())
    infos = data['hcsp_info']
    num_steps = data['num_steps']
    profile = False

    def convert_procs(procs):
        return {proc['name']:hcsp.Procedure(proc['name'], proc['text'])
                for proc in procs}

    infos = [simulator.SimInfo(info['name'], info['text'], outputs=info['outputs'],
                               procedures=convert_procs(info['procedures']))
             for info in infos if 'parallel' not in info]

    num_show = data['num_show']
    show_interval = 40000 if num_show > 40000 else None
    if 'start_event' in data:
        start_event = data['start_event']
    else:
        start_event = None

    if profile:
        pr = cProfile.Profile()
        pr.enable()

    try:
        random.seed(0)
        clock = time.perf_counter()
        res = simulator.exec_parallel(
            infos, num_steps=num_steps, num_show=num_show, show_interval=show_interval,
            start_event=start_event)
        print("Time:", time.perf_counter() - clock)

        # # export the time series data of Plant to files
        # directory = "/Users/BEAR/Projects/mars/Examples/AADL/CCS/TCS/data/"
        # s_file = open(directory + 'HCSP_s_3bus_7ms_p1s_noise_con8', 'w')
        # v_file = open(directory + 'HCSP_v_3bus_7ms_p1s_noise_con8', 'w')
        # for time_states in res['time_series']['vehicle_imp']:
        #     if 's' in time_states['state']:
        #         s_file.write(str(time_states['time']) + ',')
        #         s_file.write(str(time_states['state']['s']) + '\n')
        #     if 'v' in time_states['state']:
        #         v_file.write(str(time_states['time']) + ',')
        #         v_file.write(str(time_states['state']['v']) + '\n')
        # s_file.close()
        # v_file.close()
        #
        # obs_file = open(directory + 'HCSP_obs_3bus_7ms_p1s_noise_con8', 'w')
        # for time_states in res['time_series']['emerg_imp']:
        #     if 'obs_pos' in time_states['state']:
        #         obs_file.write(str(time_states['time']) + ',')
        #         obs_file.write(str(time_states['state']['obs_pos']) + '\n')
        # obs_file.close()
        #
        # print("DONE!")


    except simulator.SimulatorException as e:
        return raise_error(e.error_msg)

    # Process time series, so that each process has at most 10000 events
    for key in res['time_series']:
        l = len(res['time_series'][key])
        if l > 10000:
            new_series = []
            for i in range(10000):
                idx = math.floor(i * (l / 10000.0))
                new_series.append(res['time_series'][key][idx])
            res['time_series'][key] = new_series

    if len(res['trace']) < 2000:
        with open('event_output.json', 'w', encoding='utf-8') as f:
            json.dump(res['trace'], f, indent=4, ensure_ascii=False)

    if len(res['trace']) > 0:
        print_sequence_diagram(res['trace'])

    with open('simulator_events.txt', 'w', encoding='utf-8') as f:
        for i, event in enumerate(res['events']):
            f.write("%s: %s\n" % (i, event))

    if profile:
        p = Stats(pr)
        p.strip_dirs()
        p.sort_stats('cumtime')
        p.print_stats()

    for key in res.keys():
        print(key, len(json.dumps(res[key])))

    for key in res['time_series']:
        print(key, len(res['time_series'][key]))

    return json.dumps(res)

@app.route('/get_simulink_port', methods=['POST'])
def get_simulink_port():
    simulink_code = request.get_data(as_text=True)
    ports = get_simulink_port_service(simulink_code)
    return json.dumps(ports)


@app.route('/get_AADL_port', methods=['POST'])
def get_AADL_port():
    aadl_code = request.get_data(as_text=True)
    ports = get_aadl_port_service(aadl_code)
    return json.dumps(ports)


if __name__ == '__main__':
    app.run(debug=True)



import unittest

from ss2hcsp.matlab import parser
from ss2hcsp.matlab import convert


import html


class MatlabTest(unittest.TestCase):
    def testParser(self):
        s = """
        [chance(a.b)==10 && i>1]{y=chance(E,A.B);i=i+1;}/{   i=1;
            y=8;}
        """
        func = parser.transition_parser.parse(html.unescape(s))
        
        print(func.cond)
        print(func.cond_act.cmd1)

if __name__ == "__main__":
    unittest.main()


from ss2hcsp.sl.get_hcsp import new_translate_discrete, new_translate_continuous
from ss2hcsp.hcsp.expr import AConst, AVar, RelExpr, conj, disj, true_expr
from ss2hcsp.hcsp import hcsp as hp
from ss2hcsp.hcsp.module import HCSPModule


def get_thread_module(name, ports, discrete_diagram, deadline, max_et, prior=0, processor=0, reqResources=None):
    """
    :param name: the name of the thread
    :param ports: the ports of the thread in the form of {port_name: port_type}
    :param discrete_diagram: the discrete Simulink diagram of the thread
    :param deadline: the deadline of the thread, and deadline <= period
    :param max_et: maximal execution time
    :param prior: the priority of the thread
    :param processor: the thread is bound to processor
    :param reqResources: Resources required, like buses and so on
    :return: two HCSPs (modules) for the thread: Dispatch and Execution
    """

    init_hps, procedures, output_hps, update_hps, _ = new_translate_discrete(discrete_diagram)

    # Initialize the out-event port states
    init_hps.extend([hp.Assign(var_name=port_name, expr=AConst(0))
                     for port_name, port_type in ports.items() if port_type == "out event"])

    # Add state := "wait" and prior := ?? to the initialization list
    init_hps.extend([hp.Assign(var_name="state", expr=AConst("wait")),
                     hp.Assign(var_name="prior", expr=AConst(prior))])
    init_hps = hp.Sequence(*init_hps)
    # if state == "wait" then ...
    cond_wait = RelExpr("==", AVar("state"), AConst("wait"))
    hp_wait = list()
    if any(port_type == "in event" for port_type in ports.values()):
        input_events = [(hp.ParaInputChannel(ch_name="dis", paras=[name, port_name], var_name="event"), hp.Skip())
                        for port_name, port_type in ports.items() if port_type == "in event"]
        input_events = hp.SelectComm(*input_events) if len(input_events) >= 2 else input_events[0][0]
        hp_wait.append(input_events)
    else:
        hp_wait.append(hp.ParaInputChannel(ch_name="dis", paras=[name]))
    if any(port_type == "in data" for port_type in ports.values()):
        input_datas = [hp.ParaInputChannel(ch_name="inputs", paras=[name, port_name], var_name=port_name)
                       for port_name, port_type in ports.items() if port_type == "in data"]
        input_datas = hp.Sequence(*input_datas) if len(input_datas) >= 2 else input_datas[0]
        hp_wait.append(input_datas)
    hp_wait.extend([hp.Assign(var_name="t", expr=AConst(0)),
                    hp.Assign(var_name="entered", expr=AConst(0)),
                    hp.Assign(var_name="state", expr=AConst("ready"))])
    hp_wait = hp.Sequence(*hp_wait)
    # elif state == "ready" then ...
    cond_ready = RelExpr("==", AVar("state"), AConst("ready"))
    hp_ready = list()
    hp_ready.append(hp.ParaOutputChannel(ch_name="reqProcessor", paras=[processor, name], expr=AVar("prior")))
    hp_ready.append(hp.ODE_Comm(eqs=[("t", AConst(1))], constraint=RelExpr("<", AVar("t"), AConst(deadline)),
                                io_comms=[(hp.ParaInputChannel(ch_name="run", paras=[processor, name]),
                                           hp.Assign(var_name="state", expr=AConst("running")))]
                                )
                    )
    hp_ready.append(hp.Condition(cond=RelExpr("==", AVar("t"), AConst(deadline)),
                                 hp=hp.SelectComm((hp.ParaOutputChannel(ch_name="exit", paras=[processor, name]),
                                                   hp.Assign(var_name="state", expr=AConst("wait"))),
                                                  (hp.ParaInputChannel(ch_name="run", paras=[processor, name]),
                                                   hp.Assign(var_name="state", expr=AConst("running")))
                                                  )
                                 )
                    )
    hp_ready = hp.Sequence(*hp_ready)
    # elif state == "running"
    cond_running = RelExpr("==", AVar("state"), AConst("running"))
    #   entered == 0 -> ...
    entered0 = RelExpr("==", AVar("entered"), AConst(0))
    entered0_hp = list()
    entered0_hp.append(hp.Assign(var_name="c", expr=AConst(0)))
    #       The main hcsp of the thread is output_hps + updata_hps
    entered0_hp.extend([output_hp.hp for output_hp in output_hps])
    entered0_hp.extend([update_hp.hp for update_hp in update_hps])
    entered0_hp.append(hp.Assign(var_name="entered", expr=AConst(1)))
    entered0_hp = hp.Condition(cond=entered0, hp=hp.Sequence(*entered0_hp))
    #   entered == 1 -> ...
    entered1 = RelExpr("==", AVar("entered"), AConst(1))
    entered1_hp = list()
    entered1_hp.append(hp.ODE_Comm(eqs=[("t", AConst(1)), ("c", AConst(1))],
                                   constraint=conj(RelExpr("<", AVar("c"), AConst(max_et)),
                                                   RelExpr("<", AVar("t"), AConst(deadline))),
                                   io_comms=[(hp.ParaInputChannel(ch_name="preempt", paras=[processor, name]),
                                              hp.Assign(var_name="state", expr=AConst("ready")))
                                             ]
                                   )
                       )
    # state == "running" -> ...
    #   if c == max_et then ...
    cond_max_et = RelExpr("==", AVar("c"), AConst(max_et))

    _outputs = list()
    if any(port_type == "out event" for port_type in ports.values()):
        output_events = [hp.Condition(cond=RelExpr("==", AVar(port_name), AConst(1)),
                                      hp=hp.Sequence(hp.Assign(var_name=port_name, expr=AConst(0)),
                                                     hp.ParaOutputChannel(ch_name="outputs",
                                                                          paras=[name, port_name])))
                         for port_name, port_type in ports.items() if port_type == "out event"]
        output_events = hp.Sequence(*output_events) if len(output_events) >= 2 else output_events[0]
        _outputs.append(output_events)
    if any(port_type == "out data" for port_type in ports.values()):
        output_datas = [hp.ParaOutputChannel(ch_name="outputs", paras=[name, port_name], expr=AVar(port_name))
                        for port_name, port_type in ports.items() if port_type == "out data"]
        output_datas = hp.Sequence(*output_datas) if len(output_datas) >= 2 else output_datas[0]
        _outputs.append(output_datas)
    assert len(_outputs) >= 1
    _outputs = hp.Sequence(*_outputs) if len(_outputs) >= 2 else _outputs[0]

    to_wait = hp.SelectComm((hp.ParaInputChannel(ch_name="preempt", paras=[processor, name]),
                             hp.Assign(var_name="state", expr=AConst("wait"))),
                            (hp.ParaOutputChannel(ch_name="free", paras=[processor, name]),
                             hp.Assign(var_name="state", expr=AConst("wait")))
                            )
    hp_max_et = hp.Sequence(_outputs, to_wait)
    if reqResources:
        # At present, we assume that BUS is the only required resource.
        assert reqResources == "bus"
        freeResources = hp.ParaOutputChannel(ch_name="freeBus", paras=[name])
        hp_max_et = hp.Sequence(_outputs, freeResources, to_wait)
        block_hp = hp.SelectComm((hp.ParaInputChannel(ch_name="preempt", paras=[processor, name]),
                                  hp.Assign(var_name="state", expr=AConst("await"))),
                                 (hp.ParaOutputChannel(ch_name="free", paras=[processor, name]),
                                  hp.Assign(var_name="state", expr=AConst("await")))
                                 )
        hp_max_et = hp.SelectComm((hp.ParaOutputChannel(ch_name="reqBus", paras=[name]), hp_max_et),
                                  (hp.ParaInputChannel(ch_name="block", paras=[name]), block_hp))
    hp_running = hp.ITE(if_hps=[(cond_max_et, hp_max_et)], else_hp=to_wait)
    hp_running = hp.Condition(cond=cond_running, hp=hp_running)
    entered1_hp.append(hp_running)
    entered1_hp = hp.Condition(cond=entered1, hp=hp.Sequence(*entered1_hp))
    hp_running = hp.Sequence(entered0_hp, entered1_hp)
    # else (state == "await")
    unblock_hp = hp.ODE_Comm(eqs=[("t", AConst(1))], constraint=RelExpr("<", AVar("t"), AConst(deadline)),
                             io_comms=[(hp.ParaInputChannel(ch_name="unblock", paras=[name]),
                                        hp.Assign(var_name="state", expr=AConst("ready")))]
                             )
    time_out = hp.Condition(cond=RelExpr("==", AVar("t"), AConst(deadline)),
                            hp=hp.Assign(var_name="state", expr=AConst("wait"))
                            )
    hp_await = hp.Sequence(unblock_hp, time_out)

    # Get EXE_thread
    # execute_thead = hp.HCSP()
    if reqResources:
        execute_thead = hp.ITE(if_hps=[(cond_wait, hp_wait), (cond_ready, hp_ready), (cond_running, hp_running)],
                               else_hp=hp_await)
    else:
        execute_thead = hp.ITE(if_hps=[(cond_wait, hp_wait), (cond_ready, hp_ready)],
                               else_hp=hp_running)
    return HCSPModule(name=name,
                      code=hp.Sequence(init_hps, hp.Loop(execute_thead)),
                      outputs=set(port_name for port_name, port_type in ports.items() if port_type == "out data"),
                      procedures=procedures)
    # return execute_thead


def get_bus_module(name, thread_ports, device_ports, latency):
    """
    :param name: bus name
    :param thread_ports: {thread_name: {port_name: port_type (out data or out event)}}
    :param device_ports: {device_name: {port_name: port_type (out data or out event)}}
    :param latency: ms
    :return: a module BUS(name)
    """

    reqBus_hps = list()
    unblock_hps = list()
    # io_comms = list()
    for sender, ports in thread_ports.items():
        reqBus = hp.ParaInputChannel(ch_name="reqBus", paras=[sender], is_str=True)
        get_datas = list()
        put_datas = list()
        get_events = list()
        put_events = list()
        init_events = list()
        transfer = list()
        for port_name, port_type in ports.items():
            if port_type == "out data":
                get_datas.append(hp.ParaInputChannel(ch_name="outputs", paras=[sender, port_name],
                                                     var_name=port_name, is_str=True))
                put_datas.append(hp.ParaOutputChannel(ch_name="outputs", paras=[name, port_name],
                                                      expr=AVar(port_name), is_str=True))
            elif port_type == "out event":
                get_events.append(hp.ParaInputChannel(ch_name="outputs", paras=[sender, port_name],
                                                      var_name=port_name, is_str=True))
                put_events.append(hp.ParaOutputChannel(ch_name="outputs", paras=[name, port_name],
                                                       expr=AVar(port_name), is_str=True))
                init_events.append(hp.Assign(var_name=port_name, expr=AConst("")))
            else:
                raise RuntimeError("port type must be either out data or out event!")
        if get_datas:
            get_datas = hp.Sequence(*get_datas) if len(get_datas) >= 2 else get_datas[0]
            transfer.append(get_datas)
        if init_events:
            init_events = hp.Sequence(*init_events) if len(init_events) >= 2 else init_events[0]
            transfer.append(init_events)
            transfer.append(hp.Assign(var_name="_freeBus", expr=AConst(0)))
        if get_events:
            io_comms = list()
            io_comms.extend([(event, hp.Skip()) for event in get_events])
            io_comms.append((hp.ParaInputChannel(ch_name="freeBus", paras=[sender]),
                             hp.Assign(var_name="_freeBus", expr=AConst(1))))
            get_events = hp.SelectComm(*io_comms)
            loop_cond = RelExpr("==", AVar("_freeBus"), AConst(0))
            get_events = hp.Loop(constraint=loop_cond, hp=get_events)
            transfer.append(get_events)
        transfer.append(hp.Var("BLOCK_by_"+sender))
        if put_events:
            cond_hps = list()
            cond_hps.extend([hp.Condition(cond=RelExpr("!=", event.expr, AConst("")), hp=event)
                             for event in put_events])
            put_events = hp.Sequence(*cond_hps) if len(cond_hps) >= 2 else cond_hps[0]
            transfer.append(put_events)
        if put_datas:
            put_datas = hp.Sequence(*put_datas) if len(put_datas) >= 2 else put_datas[0]
            transfer.append(put_datas)
        assert len(transfer) >= 2
        transfer = hp.Sequence(*transfer)
        reqBus_hps.append((reqBus, transfer))
        unblock_hps.append((hp.ParaOutputChannel(ch_name="unblock", paras=[sender], is_str=True), hp.Skip()))

    for sender, ports in device_ports.items():
        get_datas = list()
        put_datas = list()
        get_events = list()
        for port_name, port_type in ports.items():
            if port_type == "out data":
                get_datas.append(hp.ParaInputChannel(ch_name="outputs", paras=[sender, port_name],
                                                     var_name=port_name, is_str=True))
                put_datas.append(hp.ParaOutputChannel(ch_name="outputs", paras=[name, port_name],
                                                      expr=AVar(port_name), is_str=True))
            elif port_type == "out event":
                get_events.append(hp.ParaInputChannel(ch_name="inputs", paras=[name, port_name],
                                                      var_name="event", is_str=True))
            else:
                raise RuntimeError("port type must be either out data or out event!")
        if get_datas:
            reqBus = get_datas[0]
            transfer = list()
            transfer.extend(get_datas[1:])
            transfer.append(hp.Var("BLOCK"))
            transfer.extend(put_datas)
            transfer = hp.Sequence(*transfer) if len(transfer) >= 2 else transfer[0]
            reqBus_hps.append((reqBus, transfer))
        if get_events:
            for get_event in get_events:
                put_event = hp.ParaOutputChannel(ch_name="outputs", paras=get_event.paras,
                                                 expr=get_event.var_name, is_str=True)
                transfer = hp.Sequence(hp.Var("BLOCK"), put_event)
                reqBus_hps.append((get_event, transfer))
    io_comms = list()
    io_comms.extend(reqBus_hps)
    io_comms.extend(unblock_hps)
    assert len(io_comms) >= 1
    bus_hcsp = hp.Loop(hp.SelectComm(*io_comms)) if len(io_comms) >= 2\
        else hp.Loop(hp.Sequence(io_comms[0][0], io_comms[0][1]))

    # Get procedures of BLOCK and BLOCK_by_thread
    procesures = list()
    io_comms = [(hp.ParaOutputChannel(ch_name="block", paras=[thread], is_str=True), hp.Skip())
                for thread in thread_ports.keys()]
    reset_t = hp.Assign(var_name="t", expr=AConst(0))
    if len(io_comms) >= 1:
        eqs = [("t", AConst(1))]
        constraint = RelExpr("<", AVar("t"), AConst(latency))
        ode = hp.ODE_Comm(eqs=eqs, constraint=constraint, io_comms=io_comms)
        ode_loop = hp.Loop(constraint=constraint, hp=ode)
        procesures.append(hp.Procedure(name="BLOCK", hp=hp.Sequence(reset_t, ode_loop)))
    else:
        procesures.append(hp.Procedure(name="BLOCK", hp=hp.Wait(AConst(latency))))
    for i in range(len(io_comms)):
        thread = io_comms[i][0].paras[0]
        if len(io_comms) >= 2:
            ode = hp.ODE_Comm(eqs=eqs, constraint=constraint, io_comms=io_comms[:i]+io_comms[i+1:])
            ode_loop = hp.Loop(constraint=constraint, hp=ode)
            procesures.append(hp.Procedure(name="BLOCK_by_"+thread, hp=hp.Sequence(reset_t, ode_loop)))
        else:
            procesures.append(hp.Procedure(name="BLOCK_by_"+thread, hp=hp.Wait(AConst(latency))))

    return HCSPModule(name="Bus_"+name, code=bus_hcsp, procedures=procesures)


def get_databuffer_module(recv_num=1):
    init_hp = hp.Assign(var_name="data", expr=AVar("init_value"))
    paras = ["send", "out_port"]
    io_comms = [(hp.ParaInputChannel(ch_name="outputs", paras=paras[-2:], var_name="data"), hp.Skip())]
    for i in range(recv_num):
        paras.append("recv"+str(i))
        paras.append("in_port"+str(i))
        io_comms.append((hp.ParaOutputChannel(ch_name="inputs", paras=paras[-2:], expr=AVar("data")), hp.Skip()))
    paras.append("init_value")
    transfer = hp.Loop(hp.ODE_Comm(eqs=[("data", AConst(0))], io_comms=io_comms, constraint=true_expr))

    return HCSPModule(name="DataBuffer"+str(recv_num), params=paras, code=hp.Sequence(init_hp, transfer))


def get_continuous_module(name, ports, continuous_diagram, outputs):
    """
    ports: {port_name: (var_name, port_type)}
    """
    init_hps, equations, constraints, _, _ = new_translate_continuous(continuous_diagram)
    assert isinstance(init_hps[0], hp.Assign) and init_hps[0].var_name.name == "tt" and init_hps[0].expr.value == 0
    assert equations[0][0] == "tt" and equations[0][1].value == 1
    init_hps = init_hps[1:]
    equations = equations[1:]

    in_comms = list()
    out_comms = list()
    recv_flags = list()
    send_flags = list()
    for port_name, (var_name, port_type) in ports.items():
        if port_type == "in data":
            recv_flags.append(name + "_" + port_name)
            init_hps.append(hp.Assign(var_name=recv_flags[-1], expr=AConst(0)))
            in_comms.append((hp.ParaInputChannel(ch_name="outputs", paras=[name, port_name], var_name=var_name,
                                                 is_str=True),
                             hp.Assign(var_name=recv_flags[-1], expr=AConst(1))))
        elif port_type == "out data":
            send_flags.append(name + "_" + port_name)
            init_hps.append(hp.Assign(var_name=send_flags[-1], expr=AConst(0)))
            out_comms.append((hp.ParaOutputChannel(ch_name="inputs", paras=[name, port_name], expr=AVar(var_name),
                                                   is_str=True),
                              hp.Assign(var_name=send_flags[-1], expr=AConst(1))))
        else:
            raise RuntimeError("Not implemented!")
    init_hps = hp.Sequence(*init_hps) if len(init_hps) >= 2 else init_hps[0]

    assert send_flags
    assert len(send_flags) == len(out_comms)
    send_hp = out_comms[0][0]
    if len(send_flags) >= 2:
        loop_conds = [RelExpr("==", AVar(send_flag), AConst(0)) for send_flag in send_flags]
        loop_conds = disj(*loop_conds)
        send_hp = hp.Loop(constraint=loop_conds, hp=hp.SelectComm(*out_comms))

    assert recv_flags
    assert len(recv_flags) == len(in_comms)
    recv_hp = in_comms[0][0]
    if len(recv_flags) >= 2:
        loop_conds = [RelExpr("==", AVar(recv_flag), AConst(0)) for recv_flag in recv_flags]
        loop_conds = disj(*loop_conds)
        recv_hp = hp.Loop(constraint=loop_conds, hp=hp.SelectComm(*in_comms))

    io_comms = list()
    io_comms.extend([(in_comm, hp.Skip()) for in_comm, _ in in_comms])
    io_comms.extend([(out_comm, hp.Skip()) for out_comm, _ in out_comms])
    ode = hp.ODE_Comm(eqs=equations, constraint=conj(*constraints), io_comms=io_comms)
    ode_loop = hp.Loop(ode)

    code = hp.Sequence(init_hps, send_hp, recv_hp, ode_loop)

    return HCSPModule(name=name, code=code, outputs=outputs)



# Unit tests for modules

import unittest

from ss2hcsp.hcsp.expr import AConst
from ss2hcsp.hcsp.parser import hp_parser, module_parser, system_parser, decls_parser
from ss2hcsp.hcsp.module import HCSPModule, HCSPModuleInst, HCSPSystem, HCSPDeclarations, read_file
from ss2hcsp.hcsp.hcsp import Channel


class ModuleTest(unittest.TestCase):
    def testParseModule(self):
        mod = module_parser.parse("""
            module P0():
                output x;
                begin
                    x := 0;
                    { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x; }*
                end
            endmodule
        """)

        self.assertEqual(str(mod), "P0()")
        self.assertEqual(mod.outputs, (('x',),))

    def testParseModule2(self):
        mod = module_parser.parse("""
            module P1():
                begin
                    { wait(2); p2c?x; c2p!x-1; }*
                end
            endmodule
        """)

        self.assertEqual(str(mod), "P1()")
        self.assertEqual(mod.outputs, ())

    def testParseSystem(self):
        sys = system_parser.parse("""
            system
                P0() ||
                P1()
            endsystem
        """)
        self.assertEqual(str(sys), "P0() ||\nP1()")
        self.assertEqual(repr(sys), "System(P0(); P1())")

    def testParseDecls(self):
        decls = decls_parser.parse("""
            %type: module

            module P0():
                output x;
                begin
                    x := 0;
                    {{x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x;}*
                end
            endmodule

            module P1():
                begin
                    { wait(2); p2c?x; c2p!x-1; }*
                end
            endmodule

            system
                P0() ||
                P1()
            endsystem
        """)

        self.assertEqual(str(decls), "P0()\nP1()\nP0() ||\nP1()")
        self.assertEqual(repr(decls), "Decls(\n  Module(P0)\n  Module(P1)\n  System(P0(); P1())\n)")

    def testParseModule3(self):
        mod = module_parser.parse("""
            module P0(p2c,c2p):
            output x;
            begin
                x := 0;
                { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x; }*
            end
            endmodule
        """)

        self.assertEqual(str(mod), "P0(p2c,c2p)")
        self.assertEqual(repr(mod), "Module(P0,p2c,c2p)")

    def testParseDecls2(self):
        decls = decls_parser.parse("""
            %type: module

            module P0(p2c,c2p):
            output x;
            begin
                x := 0;
                { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x; }*
            end
            endmodule

            module P1(p2c,c2p):
            begin
                { wait(2); p2c?x; c2p!x-1; }*
            end
            endmodule

            system
                P0a = P0(ch1,ch2) ||
                P1a = P1(ch1,ch2) ||
                P0b = P0(ch3,ch4) ||
                P1b = P1(ch3,ch4)
            endsystem
        """)

        self.assertEqual(str(decls),
            "P0(p2c,c2p)\nP1(p2c,c2p)\nP0a = P0(ch1, ch2) ||\nP1a = P1(ch1, ch2) ||\nP0b = P0(ch3, ch4) ||\nP1b = P1(ch3, ch4)")
        self.assertEqual(repr(decls),
            "Decls(\n  Module(P0,p2c,c2p)\n  Module(P1,p2c,c2p)\n  System(P0a=P0(ch1,ch2); P1a=P1(ch1,ch2); P0b=P0(ch3,ch4); P1b=P1(ch3,ch4))\n)")

    def testGenerateHCSPInfo(self):
        decls = HCSPDeclarations([
            HCSPModule("P0", "x := 0; { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x; }*",
                       params=("p2c", "c2p"), outputs=(("x",),)),
            HCSPModule("P1", "{ wait(2); p2c?x; c2p!x-1; }*",
                       params=("p2c", "c2p")),
            HCSPSystem([
                HCSPModuleInst("P0a", "P0", (Channel("ch1"), Channel("ch2"))),
                HCSPModuleInst("P1a", "P1", (Channel("ch1"), Channel("ch2"))),
                HCSPModuleInst("P0b", "P0", (Channel("ch3"), Channel("ch4"))),
                HCSPModuleInst("P1b", "P1", (Channel("ch3"), Channel("ch4"))),
            ])
        ])

        infos = decls.generateHCSPInfo()
        self.assertEqual(str(infos[0].hp), "x := 0; {{x_dot = 1 & true} |> [] (ch1!x --> skip;) ch2?x;}*")
        self.assertEqual(str(infos[1].hp), "{wait(2); ch1?x; ch2!x - 1;}*")
        self.assertEqual(str(infos[2].hp), "x := 0; {{x_dot = 1 & true} |> [] (ch3!x --> skip;) ch4?x;}*")
        self.assertEqual(str(infos[3].hp), "{wait(2); ch3?x; ch4!x - 1;}*")

    def testParseSystem2(self):
        sys = system_parser.parse("""
            system
                P0(ch1, ch2, $0) ||
                P1(ch1, ch2)
            endsystem
        """)

        self.assertEqual(str(sys), "P0(ch1, ch2, $0) ||\nP1(ch1, ch2)")
        self.assertEqual(repr(sys), "System(P0(ch1,ch2,0); P1(ch1,ch2))")

    def testGenerateHCSPInfo2(self):
        decls = HCSPDeclarations([
            HCSPModule("P0", "x := init_x; {{x_dot = slope & true} |> [](p2c!x --> skip;) c2p?x;}*",
                       params=("p2c", "c2p", "init_x", "slope"), outputs=(("x",),),),
            HCSPModule("P1", "{wait(dly); p2c?x; c2p!x-1;}*",
                       params=("p2c", "c2p", "dly")),
            HCSPSystem([
                HCSPModuleInst("P0a", "P0", (Channel("ch1"), Channel("ch2"), AConst(0), AConst(1))),
                HCSPModuleInst("P1a", "P1", (Channel("ch1"), Channel("ch2"), AConst(2))),
                HCSPModuleInst("P0b", "P0", (Channel("ch3"), Channel("ch4"), AConst(1), AConst(2))),
                HCSPModuleInst("P1b", "P1", (Channel("ch3"), Channel("ch4"), AConst(3))),
            ])
        ])

        infos = decls.generateHCSPInfo()
        self.assertEqual(str(infos[0].hp), "x := 0; {{x_dot = 1 & true} |> [] (ch1!x --> skip;) ch2?x;}*")
        self.assertEqual(str(infos[1].hp), "{wait(2); ch1?x; ch2!x - 1;}*")
        self.assertEqual(str(infos[2].hp), "x := 1; {{x_dot = 2 & true} |> [] (ch3!x --> skip;) ch4?x;}*")
        self.assertEqual(str(infos[3].hp), "{wait(3); ch3?x; ch4!x - 1;}*")

    def testReadFile(self):
        decls = decls_parser.parse(read_file('channels.txt'))
        self.assertEqual(str(decls), "asyncChannel(ch_out,ch_in)\nNone")

    def testParseDecls3(self):
        decls = decls_parser.parse("""
            %type: module

            import channels

            module input(ch_in):
            begin
            {
                ch_in?x;
            }*
            end
            endmodule

            module output(ch_out):
            begin
            x := 0;
            {
                x := x + 1;
                ch_out!x;
            }*
            end
            endmodule

            system
                input(ch_in) ||
                output(ch_out) ||
                asyncChannel(ch_out, ch_in)
            endsystem
        """)

        infos = decls.generateHCSPInfo()
        self.assertEqual(str(infos[0].hp), "{ch_in?x;}*")
        self.assertEqual(str(infos[1].hp), "x := 0; {x := x + 1; ch_out!x;}*")
        self.assertEqual(str(infos[2].hp), "{ch_out?x --> skip; $ ch_in!x --> skip;}*")


if __name__ == "__main__":
    unittest.main()



from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.expr import AVar, AConst, RelExpr, OpExpr, IfExpr
from ss2hcsp.hcsp import hcsp as hp


class Switch(SL_Block):
    """Switch of two dest lines."""
    def __init__(self, name, relation, threshold, st=-1):
        super(Switch, self).__init__("switch", name, 1, 3, st)

        assert relation in ('>=', '>', '<=', '<')
        self.relation = relation

        assert isinstance(threshold, (int, float))
        self.threshold = threshold

    def get_expr(self):
        in_vars = [AVar(line.name) for line in self.dest_lines]
        return IfExpr(RelExpr(self.relation, in_vars[1], AConst(self.threshold)),
                      in_vars[0], in_vars[2])

    def __str__(self):
        out_var = self.src_lines[0][0].name
        expr = self.get_expr()
        return "%s: %s = %s  (st = %s)" % (self.name, out_var, expr, self.st)

    def __repr__(self):
        return "Switch(%s, %s, %s, %s, in = %s, out = %s)" % \
            (self.name, self.relation, self.threshold, self.st, str(self.dest_lines), str(self.src_lines))

    def get_output_hp(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return hp.Assign(out_var, expr)

    def get_var_subst(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return {out_var: expr}

    def get_var_map(self):
        in_vars = [AVar(line.name) for line in self.dest_lines]
        cond0 = RelExpr(self.relation, in_vars[1], AConst(self.threshold))
        expr0 = in_vars[0]
        cond2 = cond0.neg()
        expr2 = in_vars[2]
        out_var = self.src_lines[0][0].name
        return {out_var: [(cond0, expr0), (cond2, expr2)]}



"""Unit Test for SymPy wrapper"""
import unittest

from ss2hcsp.hcsp.parser import expr_parser, expr_parser
from hhlpy.sympy_wrapper import sp_simplify, sp_polynomial_div, sp_is_polynomial

class SympyWrapperTest(unittest.TestCase):
    def testSimplifyBexpr(self):
        test_case_bexpr = {
            "~(x > 1)"          :"x <= 1",
            "~(~(x > 1))"       :"x > 1",
            "x > 1 || x > 0"    :"x > 0",
            "x > 1 && x > 2"    :"x > 2",
            "x > 1 && x > 2 && x > 3"   :"x > 3",
            "x > 2 --> 2 < x"   :"true",
            "x > 2 <--> 2 < x"  :"true"
        }
        for k, e in test_case_bexpr.items():
            k = expr_parser.parse(k)
            e = expr_parser.parse(e)

            self.assertTrue(sp_simplify(k) == e)

    def testSimplifyAexpr(self):
        test_case_aexpr = {
            "(x * y + z * y) / y"       :"x + z",
            "x ^ y - x ^ y"             :"0"
        }
        for k, e in test_case_aexpr.items():
            k = expr_parser.parse(k)
            e = expr_parser.parse(e)

            self.assertTrue(sp_simplify(k) == e)

    def testSpPolynomialDiv(self):
        test_case = {
            ("x^2 + 1", "x")        :("x", "1"),
            ("x^3 + 1", "x^2 - 1")  :("x", "1 + x")
        }

        for k, e in test_case.items():
            p = expr_parser.parse(k[0])
            q = expr_parser.parse(k[1])
            e0 = expr_parser.parse(e[0])
            e1 = expr_parser.parse(e[1])

            quot_remains = sp_polynomial_div(p, q)
            self.assertTrue(quot_remains[e0] == e1)

    def testSpIsPolynomial(self):
        test_cases = [
            ("x^3 - 2*x/y + 3*x*z", {"y"}, True),
            ("x^3 - 2*x/y + 3*x*z", set(), False),
            # ("x^2 + a*x*y^2 - b*Sin[c]", {"a", "b", "c"}): True,
            # ("x^2 + a*x*y^2 - b*Sin[c]", {"a", "b"}): False
        ]

        for case in test_cases:
            e = case[0]
            constants = case[1]
            result = case[2]

            e = expr_parser.parse(e)
           
            self.assertTrue(sp_is_polynomial(e, constants) == result)


"""Expressions"""

import math
from decimal import Decimal
from fractions import Fraction
import itertools

from ss2hcsp.util.topsort import topological_sort


def opt_round(x):
    if abs(round(x, 3) - x) < 1e-7:
        return round(x, 3)
    else:
        return x

def get_range(start, end):
    """Returns a range of numbers between start and end, inserting
    multiples of 0.1 along the way.

    """
    start_int = math.ceil(start * 10)
    end_int = math.floor(end * 10)
    res = []
    if start * 10 != start_int:
        res.append(start)
    for i in range(start_int, end_int+1):
        res.append(i / 10)
    if end * 10 != end_int:
        res.append(end)
    return res

def str_of_val(val):
    """Print form of a string."""
    if val is None:
        return ""
    elif isinstance(val, float):
        return str(round(val, 3))
    elif isinstance(val, dict):
        return "{" + (','.join(k + ':' + str(v) for k, v in val.items())) + "}"
    elif isinstance(val, str):
        return "\"" + val + "\""
    elif isinstance(val, list):
        return "[" + ",".join(str_of_val(v) for v in val) + "]"
    else:
        return str(val)


class Expr:
    """Arithmetic expression."""
    def __init__(self):
        pass

    def priority(self):
        """Returns priority during printing."""
        raise NotImplementedError

    def get_vars(self):
        """Returns set of variables in the expression."""
        raise NotImplementedError

    def get_fun_names(self):
        """Return set of function names in the expression"""
        return NotImplementedError

    def get_zero_arity_funcs(self):
        """Return set of functions with zero arity in the expression"""
        return NotImplementedError

    def subst(self, inst):
        """inst is a dictionary mapping variable names to expressions."""
        raise NotImplementedError


class AVar(Expr):
    def __init__(self, name, meta=None):
        super(AVar, self).__init__()
        assert isinstance(name, str)
        self.name = name
        self.meta = meta

    def __repr__(self):
        return "AVar(%s)" % self.name

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return isinstance(other, AVar) and self.name == other.name

    def __hash__(self):
        return hash(("AVar", self.name))

    def priority(self):
        return 100

    def get_vars(self):
        return {self.name}

    def get_fun_names(self):
        return set()

    def get_zero_arity_funcs(self):
        return set()

    def subst(self, inst):
        if self.name in inst:
            return inst[self.name]
        else:
            return self


class AConst(Expr):
    def __init__(self, value, meta=None):
        super(AConst, self).__init__()
        assert isinstance(value, (int, float, Decimal, Fraction, list, str, tuple, dict))
        # if isinstance(value, Decimal):
        #     value = float(value)
        self.value = value
        self.meta = meta

    def __repr__(self):
        return "AConst(%s)" % str_of_val(self.value)

    def __str__(self):
        return str_of_val(self.value)

    def __eq__(self, other):
        return isinstance(other, AConst) and self.value == other.value

    def __hash__(self):
        return hash(("AConst", str(self.value)))

    def priority(self):
        return 100

    def get_vars(self):
        return set()

    def get_fun_names(self):
        return set()

    def get_zero_arity_funcs(self):
        return set()

    def subst(self, inst):
        return self


class OpExpr(Expr):
    def __init__(self, op, *exprs, meta=None):
        super(OpExpr, self).__init__()
        assert op in ('+', '-', '*', '/', '%', '^')
        for expr in exprs:
            if not isinstance(expr, Expr):
                raise AssertionError
        # assert all(isinstance(expr, Expr) for expr in exprs)
        assert len(exprs) > 0 and len(exprs) <= 2, \
            "OpExpr: wrong number of arguments for %s" % op
        if len(exprs) == 1:
            assert op == '-', 'OpExpr: wrong number of arguments for %s' % op
        self.op = op
        self.exprs = tuple(exprs)
        self.meta = meta

    def __repr__(self):
        return "OpExpr(%s,%s)" % (repr(self.op), repr(self.exprs))
    
    def __str__(self):
        if len(self.exprs) == 1:
            assert self.op == '-'
            s = str(self.exprs[0])
            if self.exprs[0].priority() < self.priority():
                s = '(' + s + ')'
            return '-' + s
        else:
            s1, s2 = str(self.exprs[0]), str(self.exprs[1])
            if self.exprs[0].priority() < self.priority():
                s1 = '(' + s1 + ')'
            if self.exprs[1].priority() <= self.priority():
                s2 = '(' + s2 + ')'
            if self.op == '^':
                return s1 + self.op + s2
            else:
                return s1 + ' ' + self.op + ' ' + s2

    def __eq__(self, other):
        return isinstance(other, OpExpr) and self.op == other.op and self.exprs == other.exprs

    def __hash__(self):
        return hash(("OpExpr", self.op, self.exprs))

    def priority(self):
        if len(self.exprs) == 1:
            return 80
        elif self.op == '^':
            return 85
        elif self.op == '*' or self.op == '/' or self.op == '%':
            return 70
        else:
            return 65

    def get_vars(self):
        return set().union(*(expr.get_vars() for expr in self.exprs))

    def get_fun_names(self):
        return set().union(*(expr.get_fun_names() for expr in self.exprs))

    def get_zero_arity_funcs(self):
        return set().union(*(expr.get_zero_arity_funcs() for expr in self.exprs))

    def subst(self, inst):
        return OpExpr(self.op, *(expr.subst(inst) for expr in self.exprs))

def list_add(*args):
    if len(args) == 0:
        return 0
    elif len(args) == 1:
        return args[0]
    else:
        return OpExpr('+', args[0], list_add(*args[1:]))

def list_mul(*args):
    if len(args) == 0:
        return 1
    elif len(args) == 1:
        return args[0]
    else:
        return OpExpr('*', args[0], list_mul(*args[1:]))

class FunExpr(Expr):
    def __init__(self, fun_name, exprs, meta=None):
        super(FunExpr, self).__init__()
        self.fun_name = fun_name
        assert isinstance(self.fun_name, str)
        exprs = tuple(exprs)
        assert all(isinstance(expr, Expr) for expr in exprs)
        self.exprs = exprs
        self.meta = meta

    def __repr__(self):
        return "Fun(%s, %s)" % (self.fun_name, ", ".join(repr(expr) for expr in self.exprs))

    def __str__(self):
        return "%s(%s)" % (self.fun_name, ",".join(str(expr) for expr in self.exprs))

    def __eq__(self, other):
        return isinstance(other, FunExpr) and self.fun_name == other.fun_name and \
               self.exprs == other.exprs

    def __hash__(self):
        return hash(("Fun", self.fun_name, self.exprs))

    def priority(self):
        return 100

    def get_vars(self):
        return set().union(*(expr.get_vars() for expr in self.exprs))

    def get_fun_names(self):
        func_names = set((self.fun_name,))
        return func_names.union(*(expr.get_fun_names() for expr in self.exprs))

    def get_zero_arity_funcs(self):
        if len(self.exprs) == 0:
            zero_arity_funcs = set((self.fun_name + '(' + ')',))
        else:
            zero_arity_funcs = set()
        return zero_arity_funcs.union(*(expr.get_zero_arity_funcs() for expr in self.exprs))

    def subst(self, inst):
        return FunExpr(self.fun_name, [expr.subst(inst) for expr in self.exprs])

def replace_function(e: FunExpr, funcs=dict()):
    """Replace a FunExpr by the expr of its corresponding Function object.
    For example,
    For FunExpr bar(y), with funcs = {'bar': Function('bar', ['x'], 'x + 1')}),
    return Expr y + 1
    """
    assert e.fun_name in funcs, \
        "Function {f} is not defined".format(f=e.fun_name)
    fun_obj = funcs[e.fun_name]
    fun_obj_expr = fun_obj.expr
    assert len(e.exprs) == len(fun_obj.vars)
    length = len(e.exprs)
    for i in range(length):
        # Substitute the parameters by arguments. 
        # Example: for bar(y), substitute x with y in x + 1.
        if str(fun_obj.vars[i]) != str(e.exprs[i]):
            fun_obj_expr = fun_obj_expr.subst({fun_obj.vars[i]: e.exprs[i]})

    return fun_obj_expr


class IfExpr(Expr):
    def __init__(self, cond, expr1, expr2, meta=None):
        super(IfExpr, self).__init__()
        assert isinstance(cond, Expr) and isinstance(expr1, Expr) and isinstance(expr2, Expr)
        self.cond = cond
        self.expr1 = expr1
        self.expr2 = expr2
        self.meta = meta
    
    def __repr__(self):
        return "IfExpr(%s,%s,%s)" % (repr(self.cond), repr(self.expr1), repr(self.expr2))
    
    def __str__(self):
        s0, s1, s2 = str(self.cond), str(self.expr1), str(self.expr2)
        return "(if %s then %s else %s)" % (s0, s1, s2)

    def __eq__(self, other):
        return isinstance(other, IfExpr) and self.cond == other.cond and \
            self.expr1 == other.expr1 and self.expr2 == other.expr2

    def __hash__(self):
        return hash(("IfExpr", self.cond, self.expr1, self.expr2))

    def priority(self):
        return 100

    def get_vars(self):
        return set().union(*(arg.get_vars() for arg in [self.cond, self.expr1, self.expr2]))

    def get_fun_names(self):
        return set().union(*(arg.get_fun_names() for arg in [self.cond, self.expr1, self.expr2]))

    def get_zero_arity_funcs(self):
        return set().union(*(arg.get_zero_arity_funcs() for arg in [self.cond, self.expr1, self.expr2]))
    
    def subst(self, inst):
        return IfExpr(self.cond.subst(inst), self.expr1.subst(inst), self.expr2.subst(inst))


class ListExpr(Expr):
    """List expressions."""
    def __init__(self, *args, meta=None):
        super(ListExpr, self).__init__()
        args = tuple(args)
        assert all(isinstance(arg, Expr) for arg in args)
        self.args = args
        self.count = 0
        self.meta = meta

    def __repr__(self):
        return "List(%s)" % (','.join(repr(arg) for arg in self.args))

    def __str__(self):
        return "[%s]" % (','.join(str(arg) for arg in self.args))

    def __eq__(self, other):
        return isinstance(other, ListExpr) and self.args == other.args

    def __hash__(self):
        return hash(("ListExpr", self.args))

    def priority(self):
        return 100

    def get_vars(self):
        return set().union(*(arg.get_vars() for arg in self.args))

    def get_fun_names(self):
        return set().union(*(arg.get_fun_names() for arg in self.args))

    def get_zero_arity_funcs(self):
        return set().union(*(arg.get_zero_arity_funcs() for arg in self.args))

    def subst(self, inst):
        return ListExpr(*(expr.subst(inst) for expr in self.args))


class DictExpr(Expr):
    """Dictionary expressions (structures)."""
    def __init__(self, *args, meta=None):
        """Argument is a list of key-value pairs. Each key should be a string."""
        super(DictExpr, self).__init__()
        self.dict = dict()
        args = tuple(args)
        assert all(isinstance(k, str) and isinstance(v, Expr) for k, v in args)
        for k, v in args:
            self.dict[k] = v
        self.meta = meta

    def __repr__(self):
        return "Dict(%s)" % (','.join(k + ':' + repr(v) for k, v in self.dict.items()))

    def __str__(self):
        return "{%s}" % (','.join(k + ':' + str(v) for k, v in self.dict.items()))

    def __eq__(self, other):
        return isinstance(other, DictExpr) and self.dict == other.dict

    def __hash__(self):
        return hash(("Dict", tuple((k, v) for k, v in self.dict.items())))

    def priority(self):
        return 100

    def get_vars(self):
        return set().union(*(v.get_vars() for k, v in self.dict.items()))

    def get_fun_names(self):
        return set().union(*(v.get_fun_names() for k, v in self.dict.items()))

    def get_zero_arity_funcs(self):
        return set().union(*(v.get_zero_arity_funcs() for k, v in self.dict.items()))

    def subst(self, inst):
        return DictExpr(*((k, v.subst(inst)) for k, v in self.dict.items()))


class ArrayIdxExpr(Expr):
    """Expressions of the form a[i], where a evaluates to a list and i
    evaluates to an integer. Constructor also supports the case where the
    second argument is a list.
    
    """
    def __init__(self, expr1, expr2, meta=None):
        super(ArrayIdxExpr, self).__init__()
        if isinstance(expr1, str):
            expr1 = AVar(expr1)
        assert isinstance(expr1, Expr)
        if isinstance(expr2, Expr):
            self.expr1 = expr1
            self.expr2 = expr2
        else:
            assert isinstance(expr2, (list, tuple)) and len(expr2) >= 1
            if len(expr2) == 1:
                self.expr1 = expr1
                self.expr2 = expr2[0]
            else:
                self.expr1 = ArrayIdxExpr(expr1, expr2[:-1])
                self.expr2 = expr2[-1]
        self.meta = meta

    def __repr__(self):
        return "ArrayIdxExpr(%s,%s)" % (repr(self.expr1), repr(self.expr2))

    def __str__(self):
        return "%s[%s]" % (str(self.expr1), str(self.expr2))

    def __eq__(self, other):
        return isinstance(other, ArrayIdxExpr) and self.expr1 == other.expr1 and self.expr2 == other.expr2

    def __hash__(self):
        return hash(("ArrayIdx", self.expr1, self.expr2))

    def priority(self):
        return 100

    def get_vars(self):
        return self.expr1.get_vars().union(self.expr2.get_vars())

    def get_fun_names(self):
        return self.expr1.get_fun_names().union(self.expr2.get_fun_names())

    def get_zero_arity_funcs(self):
        return self.expr1.get_zero_arity_funcs().union(self.expr2.get_zero_arity_funcs())

    def subst(self, inst):
        return ArrayIdxExpr(expr1=self.expr1.subst(inst), expr2=self.expr2.subst(inst))


class FieldNameExpr(Expr):
    """Expression of the form a.name, where a evaluates to a structure
    and name is a field name.

    """
    def __init__(self, expr, field, meta=None):
        super(FieldNameExpr, self).__init__()
        assert isinstance(expr, Expr) and isinstance(field, str)
        self.expr = expr
        self.field = field
        self.meta = meta

    def __repr__(self):
        return "FieldNameExpr(%s,%s)" % (repr(self.expr), self.field)

    def __str__(self):
        return "%s.%s" % (str(self.expr), self.field)

    def __eq__(self, other):
        return isinstance(other, FieldNameExpr) and self.expr == other.expr and \
            self.field == other.field

    def __hash__(self):
        return hash(("FieldName", self.expr, self.field))

    def priority(self):
        return 100

    def get_vars(self):
        return self.expr.get_vars()

    def get_fun_names(self):
        return self.expr.get_fun_names()

    def get_zero_arity_funcs(self):
        return self.expr.get_zero_arity_funcs()

    def subst(self, inst):
        return FieldNameExpr(expr=self.expr.subst(inst), field=self.field)

class BConst(Expr):  # Boolean expression
    def __init__(self, value, meta=None):
        super(BConst, self).__init__()
        assert isinstance(value, bool)
        self.value = value
        self.meta = meta

    def __repr__(self):
        return "BConst(%s)" % str(self.value)

    def __str__(self):
        return "true" if self.value else "false"

    def __eq__(self, other):
        return isinstance(other, BConst) and self.value == other.value

    def __hash__(self):
        return hash(("BConst", self.value))

    def priority(self):
        return 100

    def get_vars(self):
        return set()

    def get_fun_names(self):
        return set()

    def get_zero_arity_funcs(self):
        return set()

    def subst(self, inst):
        return self

true_expr = BConst(True)
false_expr = BConst(False)

class LogicExpr(Expr):
    def __init__(self, op, *exprs, meta=None):
        super(LogicExpr, self).__init__()
        assert op in ["&&", "||", "->", "<->", "!"]
        assert all(isinstance(expr, Expr) for expr in exprs), \
            "Expected Expr: {}".format(exprs)
        assert len(exprs) > 0 and len(exprs) <= 2, \
            "LogicExpr: wrong number of arguments for %s" % op
        if len(exprs) == 1:
            assert op == '!', "LogicExpr: wrong number of arguments for %s" % op
        self.op = op
        self.exprs = tuple(exprs)
        self.meta = meta

    def __repr__(self):
        return "Logic(%s,%s)" % (repr(self.op), repr(self.exprs))

    def __str__(self):
        if len(self.exprs) == 1:
            assert self.op == '!'
            s = str(self.exprs[0])
            if self.exprs[0].priority() < self.priority():
                s = '(' + s + ')'
            return '!' + s
        else:
            s1, s2 = str(self.exprs[0]), str(self.exprs[1])
            if self.exprs[0].priority() <= self.priority():
                s1 = '(' + s1 + ')'
            if self.exprs[1].priority() < self.priority():
                s2 = '(' + s2 + ')'
            return s1 + ' ' + self.op + ' ' + s2

    def __eq__(self, other):
        return isinstance(other, LogicExpr) and self.op == other.op and self.exprs == other.exprs

    def __hash__(self):
        return hash(("Logic", self.op, self.exprs))

    def priority(self):
        if self.op == '<->':
            return 25
        elif self.op == '->':
            return 20
        elif self.op == '&&':
            return 35
        elif self.op == '||':
            return 30
        elif self.op == '!':
            return 40
        else:
            raise NotImplementedError

    def get_vars(self):
        return set().union(*(expr.get_vars() for expr in self.exprs))

    def get_fun_names(self):
        return set().union(*(expr.get_fun_names() for expr in self.exprs))

    def get_zero_arity_funcs(self):
        return set().union(*(expr.get_zero_arity_funcs() for expr in self.exprs))

    def subst(self, inst):
        return LogicExpr(self.op, *(expr.subst(inst) for expr in self.exprs))


def list_conj(*args):
    if len(args) == 0:
        return true_expr
    if len(args) == 1:
        return args[0]
    return LogicExpr("&&", args[0], list_conj(*args[1:]))

def conj(*args):
    """Form the conjunction of the list of arguments.
    
    Example: conj("x > 1", "x < 3") forms "x > 1 && x < 3"

    """
    assert isinstance(args, tuple) and all(isinstance(arg, Expr) for arg in args)
    if false_expr in args:
        return false_expr
    new_args = []
    for arg in args:
        if arg != true_expr and arg not in new_args:
            new_args.append(arg)
    return list_conj(*new_args)

def split_conj(e):
    if isinstance(e, LogicExpr) and e.op == '&&':
        return split_conj(e.exprs[0]) + split_conj(e.exprs[1])
    else:
        return [e]

def list_disj(*args):
    if len(args) == 0:
        return false_expr
    if len(args) == 1:
        return args[0]
    return LogicExpr("||", args[0], list_disj(*args[1:]))

def disj(*args):
    """Form the disjunction of the list of arguments.
    
    Example: disj("x > 1", "x < 3") forms "x > 1 || x < 3"

    """
    assert isinstance(args, tuple) and all(isinstance(arg, Expr) for arg in args)
    if true_expr in args:
        return true_expr
    new_args = []
    for arg in args:
        if arg != false_expr and arg not in new_args:
            new_args.append(arg)
    return list_disj(*new_args)

def split_disj(e):
    if isinstance(e, LogicExpr) and e.op == '||':
        return [e.exprs[0]] + split_disj(e.exprs[1])
    else:
        return [e]

def imp(b1, b2):
    if b1 == false_expr or b2 == true_expr:
        return true_expr
    if b1 == true_expr:
        return b2
    return LogicExpr("->", b1, b2)
    

class RelExpr(Expr):
    neg_table = {"<": ">=", ">": "<=", "==": "!=", "!=": "==", ">=": "<", "<=": ">"}

    def __init__(self, op, expr1, expr2, meta=None):
        super(RelExpr, self).__init__()
        assert op in ["<", ">", "==", "!=", ">=", "<="]
        assert isinstance(expr1, Expr) and isinstance(expr2, Expr)
        self.op = op
        self.expr1 = expr1
        self.expr2 = expr2
        self.meta = meta

    def __repr__(self):
        return "Rel(%s, %s, %s)" % (self.op, repr(self.expr1), repr(self.expr2))

    def __str__(self):
        return str(self.expr1) + " " + self.op + " " + str(self.expr2)

    def __eq__(self, other):
        return isinstance(other, RelExpr) and self.op == other.op and \
            self.expr1 == other.expr1 and self.expr2 == other.expr2

    def __hash__(self):
        return hash(("Rel", self.op, self.expr1, self.expr2))

    def neg(self):
        if self.op == "==":
            return disj(RelExpr(">", self.expr1, self.expr2), RelExpr("<", self.expr1, self.expr2))
        return RelExpr(RelExpr.neg_table[self.op], self.expr1, self.expr2)

    def priority(self):
        return 50

    def get_vars(self):
        return self.expr1.get_vars().union(self.expr2.get_vars())

    def get_fun_names(self):
        return self.expr1.get_fun_names().union(self.expr2.get_fun_names())

    def get_zero_arity_funcs(self):
        return self.expr1.get_zero_arity_funcs().union(self.expr2.get_zero_arity_funcs())

    def subst(self, inst):
        return RelExpr(self.op, self.expr1.subst(inst), self.expr2.subst(inst))

class ExistsExpr(Expr):
    """Exists expressions"""
    def __init__(self, vars, expr, meta=None):
        assert isinstance(expr, Expr)
        if isinstance(vars, str):
            vars = AVar(vars)
        elif isinstance(vars, list):
            assert all(isinstance(var, str) for var in vars)
            vars = tuple(AVar(var) for var in vars)
        assert isinstance(vars, (AVar, tuple))
        self.vars = vars
        self.expr = expr
        self.meta = meta

    def __repr__(self):
        if isinstance(self.vars, AVar):
            return "ExistsExpr(%s, %s)" % (repr(self.vars), repr(self.expr))
        else:
            return "ExistsExpr({%s}, %s)" % ((', '.join(repr(var) for var in self.vars)), repr(self.expr))

    def __str__(self):
        if isinstance(self.vars, AVar):
            return "\\exists %s. %s" % (str(self.vars), str(self.expr))
        else:
            return "\\exists {%s}. %s" % ((', '.join(str(var) for var in self.vars)), str(self.expr))

    def __eq__(self, other):
        # Currently does not consider alpha equivalence.
        return isinstance(other, ExistsExpr) and self.vars == other.vars and self.expr == other.expr

    def __hash__(self):
        return hash(("Exists", self.vars, self.expr))

    def priority(self):
        return 10

    def get_vars(self):
        # Currently also include the bound variable.
        return self.expr.get_vars()

    def get_fun_names(self):
        return self.expr.get_fun_names()

    def get_zero_arity_funcs(self):
        return self.expr.get_zero_arity_funcs()

    def subst(self, inst):
        # Currently assume the bound variable cannot be substituded.
        assert self.vars not in inst
        return ExistsExpr(self.vars, self.expr.subst(inst))

class ForAllExpr(Expr):
    """ForAll expressions"""
    def __init__(self, vars, expr, meta=None):
        assert isinstance(expr, Expr)
        if isinstance(vars, str):
            vars = AVar(vars)
        elif isinstance(vars, list):
            assert all(isinstance(var, str) for var in vars)
            vars = tuple(AVar(var) for var in vars)
        assert isinstance(vars, (AVar, tuple))
        self.vars = vars
        self.expr = expr
        self.meta = meta

    def __repr__(self):
        if isinstance(self.vars, AVar):
            return "ForAllExpr(%s, %s)" % (repr(self.vars), repr(self.expr))
        else:
            return "ForAllExpr({%s}, %s)" % ((', '.join(repr(var) for var in self.vars)), repr(self.expr))

    def __str__(self):
        if isinstance(self.vars, AVar):
            return "ForAll %s. %s" % (str(self.vars), str(self.expr))
        else:
            return "ForAll {%s}. %s" % ((', '.join(str(var) for var in self.vars)), str(self.expr))

    def __eq__(self, other):
        # Currently does not consider alpha equivalence.
        return isinstance(other, ForAllExpr) and self.vars == other.vars and self.expr == other.expr

    def __hash__(self):
        return hash(("ForAll", self.vars, self.expr))

    def priority(self):
        return 10

    def get_vars(self):
        # Currently also include the bound variable.
        return self.expr.get_vars()

    def get_fun_names(self):
        return self.expr.get_fun_names()
    
    def get_zero_arity_funcs(self):
        return self.expr.get_zero_arity_funcs()
    
    def subst(self, inst):
        # Currently assume the bound variable cannot be substituded.
        assert self.vars not in inst
        return ForAllExpr(self.vars, self.expr.subst(inst))

def neg_expr(e):
    """Returns the negation of an expression, using deMorgan's law to move
    negation inside.

    """
    if e == true_expr:
        return false_expr
    elif e == false_expr:
        return true_expr
    elif isinstance(e, LogicExpr):
        if e.op == '&&':
            return LogicExpr('||', neg_expr(e.exprs[0]), neg_expr(e.exprs[1]))
        elif e.op == '||':
            return LogicExpr('&&', neg_expr(e.exprs[0]), neg_expr(e.exprs[1]))
        elif e.op == '->':
            return LogicExpr('&&', e.exprs[0], neg_expr(e.exprs[1]))
        elif e.op == '<->':
            return LogicExpr('<->', e.exprs[0], neg_expr(e.exprs[1]))
        elif e.op == '!':
            return e.exprs[0]
        else:
            raise NotImplementedError
    elif isinstance(e, RelExpr):
        return e.neg()
    else:
        raise NotImplementedError

def subst_all(e, inst):
    """Perform all substitutions given in inst. Detect cycles.
    
    First compute a topological sort of dependency in inst, which will
    provide the order of substitution.

    """
    # Set of all variables to be substituted
    all_vars = set(inst.keys())

    # Mapping variable to its dependencies
    dep_order = dict()
    for var in all_vars:
        dep_order[var] = list(inst[var].get_vars().intersection(all_vars))

    topo_order = topological_sort(dep_order)
    for var in reversed(topo_order):
        e = e.subst({var: inst[var]})
    return e


class Conditional_Inst:
    """A mapping from variables to lists of conditional substitutions.

    For example:
    {
        x: [(True, y + 1)],
        z: [(x >= 1, y + 1), (x < 1, y - 1)]
    }
    """
    def __init__(self):
        self.data = dict()
        self.mu_ex_conds = list()  # record a list of sets of mutually exclusive constraints

    def __repr__(self):
        def repr_pair(cond, inst):
            return "(%s, %s)" % (str(cond), str(inst))

        def repr_cond_inst(cond_inst):
            return "[" + ", ".join(repr_pair(cond, inst) for cond, inst in cond_inst) + "]"

        return "\n".join(var + ": " + repr_cond_inst(cond_inst) for var, cond_inst in self.data.items())

    def conflicting(self, conditions):
        assert isinstance(conditions, set)
        if len(conditions) == 0:
            return False  # No conflict
        elif len(conditions) == 1:
            condition = list(conditions)[0]
            if condition == false_expr:
                return True  # Conflict
            else:
                return False  # No conflict

        for cond0, cond1 in itertools.combinations_with_replacement(conditions, 2):
            if cond0 != cond1:
                for mu_ex_cond_set in self.mu_ex_conds:
                    if {cond0, cond1}.issubset(mu_ex_cond_set):
                        return True
        return False

    def add(self, var_name, cond_inst):
        """Add a new instantiation."""
        assert var_name not in self.data
        
        # Substitute using existing instantiations
        for var in self.data:
            new_cond_inst = []
            for cond, expr in cond_inst:
                if var in cond.get_vars() or var in expr.get_vars():
                    for cond_var, expr_var in self.data[var]:
                        new_cond = cond.subst({var: expr_var})
                        if self.conflicting({new_cond, cond_var}) or conj(new_cond, cond_var) == false_expr:
                            continue  # because new_cond & cond_var is False
                        else:
                            new_cond = conj(new_cond, cond_var)
                        new_expr = expr.subst({var: expr_var})
                        new_cond_inst.append((new_cond, new_expr))
                else:
                    new_cond_inst.append((cond, expr))
            cond_inst = new_cond_inst
        # Check: the conditons in cond_inst are different
        conditions = [cond for cond, _ in cond_inst]
        assert len(conditions) == len(set(conditions))

        # Merge (cond, expr) pairs with the same expression
        expr_dict = dict()
        for cond, expr in cond_inst:
            if expr not in expr_dict:
                expr_dict[expr] = []
            expr_dict[expr].append(cond)

        cond_inst = []
        for expr, conds in sorted(expr_dict.items(), key=str):
            cond_inst.append((disj(*conds), expr))

        self.data[var_name] = cond_inst

        # Extract mutually exclusive constraints
        if len(cond_inst) >= 2:
            mu_ex_cons = set(cond for cond, _ in cond_inst)
            if mu_ex_cons not in self.mu_ex_conds:
                self.mu_ex_conds.append(mu_ex_cons)



"""Scope block"""

from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp import expr
from ss2hcsp.hcsp import hcsp as hp


class Scope(SL_Block):
    """Represents an output signal"""
    def __init__(self, name, num_dest, st=-1):
        super(Scope, self).__init__("scope", name, 0, num_dest, st)

    def __str__(self):
        in_vars = [line.name for line in self.dest_lines]
        return "%s: output %s" % (self.name, ', '.join(in_vars))

    def __repr__(self):
        return "Scope(%s, %s, %s, in = %s)" % \
            (self.name, self.num_dest, self.st, str(self.dest_lines))

    def get_output_hp(self):
        return hp.Skip()

    def get_receive_hp(self):
        chs = []
        assert len(self.dest_lines) >= 1
        if len(self.dest_lines) == 1:
            line = self.dest_lines[0]
            return hp.Loop(hp.InputChannel(line.ch_name, expr.AVar(line.ch_name)))
        else:
            for line in self.dest_lines:
                chs.append((hp.InputChannel(line.ch_name, expr.AVar(line.ch_name)), hp.Skip()))
            return hp.Loop(hp.SelectComm(*chs))



# Unit test for topological sort

import unittest

from ss2hcsp.util.topsort import topological_sort, TopologicalSortException

class TopSortTest(unittest.TestCase):
    def testTopSort1(self):
        data = {
            1: {2, 3},
            2: {3, 4},
            3: {4},
            4: {}
        }
        self.assertEqual(topological_sort(data), [4, 3, 2, 1])

    def testTopSort2(self):
        data = {
            1: {2, 3},
            2: {3, 4},
            3: {4},
            4: {1}
        }
        self.assertRaises(TopologicalSortException, topological_sort, data)

    def testTopSort3(self):
        data = {
            1: {4},
            2: {4},
            3: {4},
            4: {}
        }
        self.assertEqual(topological_sort(data), [4, 1, 2, 3])


if __name__ == "__main__":
    unittest.main()



"""Unit test for wolframengine_wrapper"""

import unittest

from hhlpy.wolframengine_wrapper import wl_simplify, wl_polynomial_div, wl_is_polynomial
from hhlpy.wolframengine_wrapper import session
from ss2hcsp.hcsp import expr
from ss2hcsp.hcsp.parser import expr_parser, expr_parser

class WolframWrapperTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        try:
            session.start()
        except Exception as e:
            session.start()

    @classmethod
    def tearDownClass(cls):
        session.terminate()

    def testWlSimplify(self):
        test_case_bexpr = {
            "~(x > 1)"          :"x <= 1",
            "~(~(x > 1))"       :"x > 1",
            "x > 1 || x > 0"    :"x > 0",
            "x > 1 && x > 2"    :"x > 2",
            "x > 1 && x > 2 && x > 3"   :"x > 3",
            "x > 2 --> 2 < x"   :"true",
            "x > 2 <--> 2 < x"  :"true"
        }
        for k, e in test_case_bexpr.items():
            k = expr_parser.parse(k)
            e = expr_parser.parse(e)

            self.assertTrue(wl_simplify(k) == e)

    def testWlPolynomialDiv(self):
        test_cases = {
            ("x^2 + 1", "x")        :(("x", "1"),),
            ("x^3 + 1", "x^2 - 1")  :(("x", "1 + x"),)
        }

        for k, vals in test_cases.items():
            p = expr_parser.parse(k[0])
            q = expr_parser.parse(k[1])
            quot_remains = wl_polynomial_div(p, q)

            for val in vals:
                val0 = expr_parser.parse(val[0])
                val1 = expr_parser.parse(val[1])

                self.assertTrue(quot_remains[val0] == val1)

    def testWlIsPolynomial(self):
        test_cases = [
            ("x^3 - 2*x/y + 3*x*z", {"y"}, True),
            ("x^3 - 2*x/y + 3*x*z", set(), False),
            # ("x^2 + a*x*y^2 - b*Sin[c]", {"a", "b", "c"}): True,
            # ("x^2 + a*x*y^2 - b*Sin[c]", {"a", "b"}): False
        ]

        for case in test_cases:
            e = case[0]
            constants = case[1]
            result = case[2]

            e = expr_parser.parse(e)

            self.assertTrue(wl_is_polynomial(e, constants) == result)


if __name__ == "__main__":
    unittest.main()



from ss2hcsp.hcsp.expr import Expr
from ss2hcsp.hcsp.label import Label

class Assertion:
    """Arithmetic expression."""
    def __init__(self):
        pass

class OrdinaryAssertion(Assertion):
    """Represents an assertion with its expression and proof methods.
    expr : Expr - assertion expression
    proof_methods: tuple - methods used for proof. 

    Both loop invariant with proof methods and post condition with proof methods are
    instances of OrdinaryAssertion.
    """
    def __init__(self, expr, proof_methods, meta=None):
        super(OrdinaryAssertion, self).__init__()
        assert isinstance(expr, Expr)
        assert isinstance(proof_methods, ProofMethods)
        self.meta = meta
        self.expr = expr
        self.proof_methods = proof_methods


class CutInvariant(Assertion):
    def __init__(self, expr, proof_methods, rule = None, rule_arg=None, meta=None):
        super(CutInvariant, self).__init__()
        assert isinstance(expr, Expr)
        assert isinstance(proof_methods, ProofMethods)
        if rule:
            assert isinstance(rule, str)
        if rule_arg:
            assert isinstance(rule_arg, Expr)
        self.meta = meta
        self.expr = expr
        self.proof_methods = proof_methods
        self.rule = rule
        self.rule_arg = rule_arg

class GhostIntro(Assertion):
    def __init__(self, var, diff, meta=None):
        super(GhostIntro, self).__init__()
        self.meta = meta
        self.var = var
        self.diff = diff


class ProofMethod():
    """Represents the label and corresponding proof method.
    For example, label: 1.1
                 method: z3            
    """
    def __init__(self, method, label=None, meta=None):
        self.meta = meta
        # label can be None
        if label:
            assert isinstance(label, Label)
        assert isinstance(method, str)
        self.label = label
        self.method = method

    def __str__(self):
        if self.label:
            return str(self.label) + ': ' + self.method
        else:
            return self.method

class ProofMethods():

    def __init__(self, *pms, meta=None):
       
        assert all(isinstance(pm, ProofMethod) for pm in pms)

        self.pms = tuple(pms)
        self.meta = meta


from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.expr import AVar, AConst, OpExpr, FunExpr, true_expr


class Logic(SL_Block):
    def __init__(self, name, num_dest, st=-1):
        super(Logic, self).__init__()
        self.name = name
        self.type = "logic"
        self.is_continuous = (st == 0)
        self.num_src = 1
        self.num_dest = num_dest
        self.src_lines = [[]]
        self.dest_lines = [None] * self.num_dest

        assert isinstance(st, (int, float))
        self.st = st

    def get_var_map(self):
        in_vars = [AVar(line.name) for line in self.dest_lines]
        if isinstance(self, And):
            expr = FunExpr("min", in_vars)
        elif isinstance(self, Or):
            expr = FunExpr("max", in_vars)
        elif isinstance(self, Not):
            assert len(in_vars) == 1
            expr = OpExpr("-", AConst(1), in_vars[0])
        else:
            raise RuntimeError("Error Type!")
        out_var = self.src_lines[0][0].name
        return {out_var: [(true_expr, expr)]}


class And(Logic):
    def __init__(self, name, num_dest, st=-1):
        super(And, self).__init__(name, num_dest, st)
        self.type = "and"

    def __str__(self):
        return "%s: And[in = %s, out = %s, st = %s]" % \
               (self.name, str(self.dest_lines), str(self.src_lines), str(self.st))

    def __repr__(self):
        return str(self)


class Or(Logic):
    def __init__(self, name, num_dest, st=-1):
        super(Or, self).__init__(name, num_dest, st)
        self.type = "or"

    def __str__(self):
        return "%s: Or[in = %s, out = %s, st = %s]" % \
               (self.name, str(self.dest_lines), str(self.src_lines), str(self.st))

    def __repr__(self):
        return str(self)


class Not(Logic):
    def __init__(self, name, num_dest=1, st=-1):
        assert num_dest == 1
        super(Not, self).__init__(name, num_dest, st)
        self.type = "not"

    def __str__(self):
        return "%s: Not[in = %s, out = %s, st = %s]" % \
               (self.name, str(self.dest_lines), str(self.src_lines), str(self.st))

    def __repr__(self):
        return str(self)



from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.expr import AVar, AConst, LogicExpr, RelExpr, IfExpr, OpExpr
from ss2hcsp.hcsp import hcsp as hp


class Saturation(SL_Block):
    """Compute the saturation value of the dest_line wrt. the upper and lower limits."""
    def __init__(self, name, up_lim, low_lim, st=-1):
        super(Saturation, self).__init__("saturation", name, 1, 1, st)

        assert isinstance(up_lim, (int, float)) and isinstance(low_lim, (int, float))
        self.up_lim = up_lim
        self.low_lim = low_lim

    def get_expr(self):
        in_var = AVar(self.dest_lines[0].name)
        return IfExpr(RelExpr("<=", in_var, AConst(self.low_lim)), AConst(self.low_lim),
                      IfExpr(RelExpr(">=", in_var, AConst(self.up_lim)), AConst(self.up_lim), in_var))

    def __str__(self):
        out_var = self.src_lines[0][0].name
        expr = self.get_expr()
        return "%s: %s = %s  (st = %s)" % (self.name, out_var, expr, self.st)

    def __repr__(self):
        return "Saturation(%s, %s, %s, %s, in = %s, out = %s)" % \
            (self.name, self.up_lim, self.low_lim, self.st, str(self.dest_lines), str(self.src_lines))

    def get_output_hp(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return hp.Assign(out_var, expr)

    def get_var_subst(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return {out_var: expr}

    def get_var_map(self):
        in_var = AVar(self.dest_lines[0].name)
        cond0 = RelExpr("<", in_var, AConst(self.low_lim))
        expr0 = AConst(self.low_lim)
        cond1 = RelExpr(">", in_var, AConst(self.up_lim))
        expr1 = AConst(self.up_lim)
        cond2 = LogicExpr("&&", cond0.neg(), cond1.neg())
        expr2 = in_var
        out_var = self.src_lines[0][0].name
        return {out_var: [(cond0, expr0), (cond1, expr1), (cond2, expr2)]}



"""Unit test for sf_isabelle"""

import unittest

from ss2hcsp.sl.sl_diagram import SL_Diagram
from ss2hcsp.sf.sf_state import Junction, OR_State, AND_State
from ss2hcsp.sf.sf_isabelle import translate_composition, translate_expr, translate_action, translate_actions, \
    translate_event, translate_trans, translate_state, translate_junction_function, dfs_search_chart, \
    translate_chart_info, translate_fe_info, translate_ge_info
from ss2hcsp.matlab import function
from ss2hcsp.matlab.parser import expr_parser, cond_parser, cmd_parser, event_parser


class SFIsabelleTest(unittest.TestCase):
    '''
    def testTranslateExpr(self):
        test_data = [
            ("0", "N 0"),
            ("x + 1", "Plus (V ''x'') (N 1)"),
        ]
        for e, res in test_data:
            e = expr_parser.parse(e)
            self.assertEqual(translate_expr(e), res)

    def testTranslateCond(self):
        test_data = [
            ("true", "Bc True"),
            ("false", "Bc False"),
            ("x > 10", "(V ''x'') [>] (N 10)"),
        ]

        for e, res in test_data:
            e = cond_parser.parse(e)
            self.assertEqual(translate_expr(e), res)

    def testTranslateEvent(self):
        test_data = [
            ("e", "S [\'\'e\'\']"),
        ]

        for event, res in test_data:
            event = event_parser.parse(event)
            self.assertEqual(translate_event(event), res)
    '''
    def testTranslate1(self):
        filename = "./Examples/Stateflow/translate/States/States1.xml"
        diagram = SL_Diagram(location=filename)
        diagram.parse_xml()
        diagram.add_line_name()
        _, _, charts, _, _, _, _, _, _ = diagram.seperate_diagram()
        chart = charts[0]
        print(chart)
        print('\n')
        save_name  = filename.split('/')[-1].split('.')[0]
        str = 'theory %s\n  imports Final_ML \nbegin\n\n' %save_name
        chart_str, def_list = dfs_search_chart(chart.diagram.ssid, chart, '', [])
        str += chart_str
        #print(chart_str)

        junc_str = translate_junction_function(chart)
        str += junc_str + '\n\n'
        def_list.append('g_def')

        v_str = 'definition v :: vals where \" v = Vals (λstr. 0) (λp str. 0) (λp. 0) ([],[]) \"\n'
        str += v_str + '\n'
        def_list.append('v_def')
        
        #chart_str = translate_chart_info(chart)
        chart_str = 'definition I :: ctxt where \n\"I str = (Info False [] [])\"'
        str += chart_str + '\n'
        #print(str)
        def_list.append('I_def')

        #fe and ge
        fe_str = 'definition fe::fenv where \" ' + '\n'
        fe_str += translate_fe_info(chart)

        #print(fe_str)
        str += fe_str + '\n'
        def_list.append('fe_def')

        ge_str = 'definition ge::genv where \" ' + '\n'
        ge_str += translate_ge_info(chart)

        print(ge_str)
        str += ge_str + '\n'
        def_list.append('ge_def')

        env_str = 'definition env::env where \"env = Env Root fe ge g\" ' + '\n'
        str += env_str
        def_list.append('env_def')

        status_str = 'definition s::status where \" s = Status v I\" ' + '\n'
        str += status_str
        def_list.append('s_def')

        str += 'text‹EXECUTION PROOF›\n'

        str += 'schematic_goal \"Root_Exec_for_times env \'\'\'\' (%s::int) s ?s\"\n' %'2'

        str += '  unfolding '
        cnt = len('  unfolding ')
        for mydef in def_list:
            str += mydef + ' '
            cnt += len(mydef) + 1
            if(cnt > 80):
                str += '\n'
                cnt = 0
        #str += '\n  apply simp'
        str += '\n  by stateflow_execution2\n\n'
        str += 'end'
        str = str.replace("-", "_")
        str = str.replace("(_", "(-")
        print(str)
        save_name = "Semantic_Stateflow/" + save_name + '.thy'
        f = open(save_name, 'w')
        f.write(str)
        f.close()
'''
    def testTranslate2(self):
        test_data = [
            ("e", "S [\'\'e\'\']"),
        ]

        for event, res in test_data:
            event = event_parser.parse(event)
            self.assertEqual(translate_event(event), res)
'''

if __name__ == "__main__":
    suite = unittest.TestSuite()
    suite.addTest(SFIsabelleTest('testTranslateHp'))

    #suite =  unittest.TestLoader().loadTestsFromTestCase(SFIsabelleTest)
    unittest.TextTestRunner(verbosity=1).run(suite)



"""Label System for distinguishing verification conditions"""
class Label:
    def __init__(self):
        pass


class BranchLabel(Label):
    """Label that records the executed program branches, 
    which generate the corresponding verification conditions.
    """
    def __init__(self):
        super(BranchLabel, self).__init__()
        pass


class AtomLabel(BranchLabel):
    """Label that records which program branch executed in ITE, IChoice or ODE.
    Example: 1 or skip"""
    def __init__(self, value):
        super(AtomLabel, self).__init__()
        assert isinstance(value, (str, int))
        self.value = value

    def __str__(self):
        return str(self.value)


class SequenceLabel(BranchLabel):
    """Label that records the program branch executed in a Sequence program.
    Example: 1.2.1"""
    def __init__(self, *labels):
        super(SequenceLabel, self).__init__()
        assert all(isinstance(lb, BranchLabel) or lb is None for lb in labels) # arguments can be None
        valid_labels = [lb for lb in labels if lb]
        assert len(valid_labels)
        self.labels = tuple(valid_labels)  # Only consider the valid labels

    def __str__(self):
        return ".".join(str(lb) for lb in self.labels)


class NestLabel(BranchLabel):
    """Label that records the program branch executed in a nested program
    Example 1(2)"""
    def __init__(self, value, sub_label):
        super(NestLabel, self).__init__()
        assert isinstance(value, (int, str))
        assert isinstance(sub_label, BranchLabel)
        self.value = value
        self.sub_label = sub_label

    def __str__(self):
        return str(self.value) + "(" + str(self.sub_label) + ")"


class CategLabel(Label):
    """Label that records the category.
    Example: init"""
    def __init__(self, categ):
        super(CategLabel, self).__init__()
        assert isinstance(categ, str)
        self.categ = categ

    def __str__(self):
        return str(self.categ)


class CompLabel(Label):
    """Label that is composed of category and branch_label.
    Example: init 1.1"""
    def __init__(self, categ_label, branch_label):
        super(CompLabel, self).__init__()
        assert isinstance(categ_label, CategLabel)
        assert isinstance(branch_label, BranchLabel)
        self.categ_label = categ_label
        self.branch_label = branch_label

    def __str__(self):
        return str(self.categ_label) + ' ' + str(self.branch_label)